import os
import re
from io import IOBase
from unittest.mock import call, patch

import sh
from testfixtures import TempDirectory
from tests.helpers.base_step_implementer_test_case import \
    BaseStepImplementerTestCase
from tests.helpers.test_utils import Any, create_sh_side_effect
from ploigos_step_runner import StepResult
from ploigos_step_runner.exceptions import StepRunnerException
from ploigos_step_runner.step_implementers.deploy.argocd import ArgoCD


class TestStepImplementerDeployArgoCDBase(BaseStepImplementerTestCase):
    def create_step_implementer(
        self,
        step_config={},
        parent_work_dir_path='',
        environment=None
    ):
        return self.create_given_step_implementer(
            step_implementer=ArgoCD,
            step_config=step_config,
            step_name='deploy',
            implementer='ArgoCD',
            parent_work_dir_path=parent_work_dir_path,
            environment=environment
        )

class TestStepImplementerDeployArgoCD_Other(TestStepImplementerDeployArgoCDBase):
    def test_step_implementer_config_defaults(self):
        defaults = ArgoCD.step_implementer_config_defaults()
        expected_defaults = {
            'argocd-sync-timeout-seconds': 60,
            'argocd-sync-retry-limit': 3,
            'argocd-auto-sync': True,
            'argocd-skip-tls' : False,
            'deployment-config-helm-chart-path': './',
            'deployment-config-helm-chart-additional-values-files': [],
            'additional-helm-values-files': [],
            'deployment-config-helm-chart-values-file-image-tag-yq-path': 'image_tag',
            'force-push-tags': False,
            'kube-api-skip-tls': False,
            'kube-api-uri': 'https://kubernetes.default.svc',
            'git-name': 'Ploigos Robot'
        }
        self.assertEqual(defaults, expected_defaults)

    def test__required_config_or_result_keys(self):
        required_keys = ArgoCD._required_config_or_result_keys()
        expected_required_keys = [
            'argocd-username',
            'argocd-password',
            'argocd-api',
            'argocd-skip-tls',
            'deployment-config-repo',
            'deployment-config-helm-chart-path',
            'deployment-config-helm-chart-values-file-image-tag-yq-path',
            'git-email',
            'git-name',
            'container-image-tag'
        ]
        self.assertEqual(required_keys, expected_required_keys)

class TestStepImplementerDeployArgoCD_validate_required_config_or_previous_step_result_artifact_keys(
    TestStepImplementerDeployArgoCDBase
):
    def test_ArgoCD_validate_required_config_or_previous_step_result_artifact_keys_success_ssh(self):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            step_config = {
                'argocd-username': 'argo-username',
                'argocd-password': 'argo-password',
                'argocd-api': 'https://argo.ploigos.xyz',
                'argocd-skip-tls': False,
                'deployment-config-repo': 'git@git.ploigos.xyz:foo/deploy-config',
                'deployment-config-helm-chart-path': 'charts/foo',
                'deployment-config-helm-chart-values-file-image-tag-yq-path': 'image.tag',
                'git-email': 'git@ploigos.xyz',
                'git-name': 'Ploigos',
                'container-image-tag': 'v0.42.0'
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )

            step_implementer._validate_required_config_or_previous_step_result_artifact_keys()

    def test_ArgoCD_validate_required_config_or_previous_step_result_artifact_keys_success_https(self):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            step_config = {
                'argocd-username': 'argo-username',
                'argocd-password': 'argo-password',
                'argocd-api': 'https://argo.ploigos.xyz',
                'argocd-skip-tls': False,
                'deployment-config-repo': 'https://git.ploigos.xyz/foo/deploy-config',
                'deployment-config-helm-chart-path': 'charts/foo',
                'deployment-config-helm-chart-values-file-image-tag-yq-path': 'image.tag',
                'git-email': 'git@ploigos.xyz',
                'git-name': 'Ploigos',
                'container-image-tag': 'v0.42.0',
                'git-username': 'test-git-username',
                'git-password': 'test-secret'
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )

            step_implementer._validate_required_config_or_previous_step_result_artifact_keys()

    def test_ArgoCD_validate_required_config_or_previous_step_result_artifact_keys_fail_https_no_git_username(self):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            step_config = {
                'argocd-username': 'argo-username',
                'argocd-password': 'argo-password',
                'argocd-api': 'https://argo.ploigos.xyz',
                'argocd-skip-tls': False,
                'deployment-config-repo': 'https://git.ploigos.xyz/foo/deploy-config',
                'deployment-config-helm-chart-path': 'charts/foo',
                'deployment-config-helm-chart-values-file-image-tag-yq-path': 'image.tag',
                'git-email': 'git@ploigos.xyz',
                'git-name': 'Ploigos',
                'container-image-tag': 'v0.42.0',
                'git-password': 'test-secret'
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )

            with self.assertRaisesRegex(
                StepRunnerException,
                "Either 'git-username' or 'git-password 'is not set. Neither or both must be set."
            ):
                step_implementer._validate_required_config_or_previous_step_result_artifact_keys()

    def test_ArgoCD_validate_required_config_or_previous_step_result_artifact_keys_fail_https_no_git_password(self):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            step_config = {
                'argocd-username': 'argo-username',
                'argocd-password': 'argo-password',
                'argocd-api': 'https://argo.ploigos.xyz',
                'argocd-skip-tls': False,
                'deployment-config-repo': 'https://git.ploigos.xyz/foo/deploy-config',
                'deployment-config-helm-chart-path': 'charts/foo',
                'deployment-config-helm-chart-values-file-image-tag-yq-path': 'image.tag',
                'git-email': 'git@ploigos.xyz',
                'git-name': 'Ploigos',
                'container-image-tag': 'v0.42.0',
                'git-username': 'test-git-username'
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )

            with self.assertRaisesRegex(
                StepRunnerException,
                "Either 'git-username' or 'git-password 'is not set. Neither or both must be set."
            ):
                step_implementer._validate_required_config_or_previous_step_result_artifact_keys()

    def test_ArgoCD_validate_required_config_or_previous_step_result_artifact_keys_fail_https_no_git_creds(self):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            step_config = {
                'argocd-username': 'argo-username',
                'argocd-password': 'argo-password',
                'argocd-api': 'https://argo.ploigos.xyz',
                'argocd-skip-tls': False,
                'deployment-config-repo': 'https://git.ploigos.xyz/foo/deploy-config',
                'deployment-config-helm-chart-path': 'charts/foo',
                'deployment-config-helm-chart-values-file-image-tag-yq-path': 'image.tag',
                'git-email': 'git@ploigos.xyz',
                'git-name': 'Ploigos',
                'container-image-tag': 'v0.42.0'
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )

            with self.assertRaisesRegex(
                StepRunnerException,
                r"Since provided 'deployment-config-repo'"
                rf" \({step_config['deployment-config-repo']}\) uses"
                r" http/https protical both 'git-username' and 'git-password' must be provided."
            ):
                step_implementer._validate_required_config_or_previous_step_result_artifact_keys()

    def test_ArgoCD_validate_required_config_or_previous_step_result_artifact_keys_fail_http_no_git_creds(self):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            step_config = {
                'argocd-username': 'argo-username',
                'argocd-password': 'argo-password',
                'argocd-api': 'https://argo.ploigos.xyz',
                'argocd-skip-tls': False,
                'deployment-config-repo': 'http://git.ploigos.xyz/foo/deploy-config',
                'deployment-config-helm-chart-path': 'charts/foo',
                'deployment-config-helm-chart-values-file-image-tag-yq-path': 'image.tag',
                'git-email': 'git@ploigos.xyz',
                'git-name': 'Ploigos',
                'container-image-tag': 'v0.42.0'
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )

            with self.assertRaisesRegex(
                StepRunnerException,
                r"Since provided 'deployment-config-repo'"
                rf" \({step_config['deployment-config-repo']}\) uses"
                r" http/https protical both 'git-username' and 'git-password' must be provided."
            ):
                step_implementer._validate_required_config_or_previous_step_result_artifact_keys()

# NOTE:
#   Could definitely do some more negative testing of _run_step testing what happens when each
#   and every mocked function throws an error (that can throw an error)
class TestStepImplementerDeployArgoCD_run_step(TestStepImplementerDeployArgoCDBase):
    @patch.object(
        ArgoCD,
        '_ArgoCD__argocd_get_app_manifest',
        return_value='/does/not/matter/manifest.yaml'
    )
    @patch.object(ArgoCD, '_ArgoCD__get_app_name', return_value='test-app-name')
    @patch.object(ArgoCD, '_ArgoCD__update_yaml_file_value')
    @patch.object(ArgoCD, '_ArgoCD__get_deployment_config_repo_tag', return_value='v0.42.0')
    @patch.object(ArgoCD, '_ArgoCD__git_tag_and_push_deployment_config_repo')
    @patch.object(ArgoCD, '_ArgoCD__argocd_add_target_cluster')
    @patch.object(ArgoCD, '_ArgoCD__clone_repo', return_value='/does/not/matter')
    @patch.object(ArgoCD, '_ArgoCD__git_commit_file')
    @patch.object(ArgoCD, '_ArgoCD__argocd_sign_in')
    @patch.object(ArgoCD, '_ArgoCD__argocd_app_create_or_update')
    @patch.object(ArgoCD, '_ArgoCD__argocd_app_sync')
    @patch.object(
        ArgoCD,
        '_ArgoCD__get_deployed_host_urls',
        return_value=['https://fruits.ploigos.xyz']
    )
    @patch.object(
        ArgoCD,
        '_ArgoCD__get_deployment_config_helm_chart_environment_values_file',
        return_value='values-PROD.yaml'
    )
    @patch.object(ArgoCD, '_ArgoCD__get_repo_branch', return_value='feature/test')
    def test_ArgoCD_run_step_success(
        self,
        get_repo_branch_mock,
        get_deployment_config_helm_chart_environment_values_file_mock,
        get_deployed_host_urls_mock,
        argocd_app_sync_mock,
        argocd_app_create_or_update_mock,
        argocd_sign_in_mock,
        git_commit_file_mock,
        clone_repo_mock,
        argocd_add_target_cluster_mock,
        git_tag_and_push_deployment_config_repo_mock,
        get_deployment_config_repo_tag_mock,
        update_yaml_file_value_mock,
        get_app_name_mock,
        argocd_get_app_manifest_mock
    ):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            step_config = {
                'argocd-username': 'argo-username',
                'argocd-password': 'argo-password',
                'argocd-api': 'https://argo.ploigos.xyz',
                'argocd-skip-tls': False,
                'deployment-config-repo': 'https://git.ploigos.xyz/foo/deploy-config',
                'deployment-config-helm-chart-path': 'charts/foo',
                'deployment-config-helm-chart-values-file-image-tag-yq-path': 'image.tag',
                'git-email': 'git@ploigos.xyz',
                'git-name': 'Ploigos',
                'container-image-tag': 'v0.42.0'
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
                environment='PROD'
            )

            actual_step_results = step_implementer._run_step()
            expected_step_result = StepResult(
                step_name='deploy',
                sub_step_name='ArgoCD',
                sub_step_implementer_name='ArgoCD',
                environment='PROD'
            )
            expected_step_result.add_artifact(
                name='argocd-app-name',
                value='test-app-name'
            )
            expected_step_result.add_artifact(
                name='config-repo-git-tag',
                value='v0.42.0'
            )
            expected_step_result.add_artifact(
                name='argocd-deployed-manifest',
                value='/does/not/matter/manifest.yaml'
            )
            expected_step_result.add_artifact(
                name='deployed-host-urls',
                value=['https://fruits.ploigos.xyz']
            )
            self.assertEqual(actual_step_results, expected_step_result)

            get_repo_branch_mock.assert_called_once_with()
            get_deployment_config_helm_chart_environment_values_file_mock.assert_called_once_with()
            get_app_name_mock.assert_called_once_with()
            deployment_config_repo_dir = os.path.join(
                step_implementer.work_dir_path,
                'deployment-config-repo'
            )
            clone_repo_mock.assert_called_once_with(
                repo_dir=deployment_config_repo_dir,
                repo_url=step_config['deployment-config-repo'],
                repo_branch='feature/test',
                git_email=step_config['git-email'],
                git_name=step_config['git-name'],
                username=None,
                password=None
            )
            update_yaml_file_value_mock.assert_called_once_with(
                file='/does/not/matter/charts/foo/values-PROD.yaml',
                yq_path=step_config['deployment-config-helm-chart-values-file-image-tag-yq-path'],
                value=step_config['container-image-tag']
            )
            git_commit_file_mock.assert_called_once_with(
                git_commit_message='Updating values for deployment to PROD',
                file_path='charts/foo/values-PROD.yaml',
                repo_dir='/does/not/matter'
            )
            get_deployment_config_repo_tag_mock.assert_called_once_with()
            git_tag_and_push_deployment_config_repo_mock.assert_called_once_with(
                deployment_config_repo=step_config['deployment-config-repo'],
                deployment_config_repo_dir='/does/not/matter',
                deployment_config_repo_tag='v0.42.0',
                force_push_tags=False
            )
            argocd_sign_in_mock.assert_called_once_with(
                argocd_api=step_config['argocd-api'],
                username=step_config['argocd-username'],
                password=step_config['argocd-password'],
                insecure=step_config['argocd-skip-tls']
            )
            argocd_add_target_cluster_mock.assert_called_once_with(
                kube_api='https://kubernetes.default.svc',
                kube_api_token=None,
                kube_api_skip_tls=False
            )
            argocd_app_create_or_update_mock.assert_called_once_with(
                argocd_app_name='test-app-name',
                repo=step_config['deployment-config-repo'],
                revision='feature/test',
                path=step_config['deployment-config-helm-chart-path'],
                dest_server='https://kubernetes.default.svc',
                auto_sync=True,
                values_files=['values-PROD.yaml']
            )
            argocd_app_sync_mock.assert_called_once_with(
                argocd_app_name='test-app-name',
                argocd_sync_timeout_seconds=60,
                argocd_sync_retry_limit=3
            )
            argocd_get_app_manifest_mock.assert_called_once_with(
                argocd_app_name='test-app-name'
            )
            get_deployed_host_urls_mock.assert_called_once_with(
                manifest_path='/does/not/matter/manifest.yaml'
            )

    @patch.object(
        ArgoCD,
        '_ArgoCD__argocd_get_app_manifest',
        return_value='/does/not/matter/manifest.yaml'
    )
    @patch.object(ArgoCD, '_ArgoCD__get_app_name', return_value='test-app-name')
    @patch.object(ArgoCD, '_ArgoCD__update_yaml_file_value')
    @patch.object(ArgoCD, '_ArgoCD__get_deployment_config_repo_tag', return_value='v0.42.0')
    @patch.object(ArgoCD, '_ArgoCD__git_tag_and_push_deployment_config_repo')
    @patch.object(ArgoCD, '_ArgoCD__argocd_add_target_cluster')
    @patch.object(ArgoCD, '_ArgoCD__clone_repo', return_value='/does/not/matter')
    @patch.object(ArgoCD, '_ArgoCD__git_commit_file')
    @patch.object(ArgoCD, '_ArgoCD__argocd_sign_in')
    @patch.object(ArgoCD, '_ArgoCD__argocd_app_create_or_update')
    @patch.object(ArgoCD, '_ArgoCD__argocd_app_sync')
    @patch.object(
        ArgoCD,
        '_ArgoCD__get_deployed_host_urls',
        return_value=['https://fruits.ploigos.xyz']
    )
    @patch.object(
        ArgoCD,
        '_ArgoCD__get_deployment_config_helm_chart_environment_values_file',
        return_value='values-PROD.yaml'
    )
    @patch.object(ArgoCD, '_ArgoCD__get_repo_branch', return_value='feature/test')
    def test_ArgoCD_run_step_success_additional_helm_values_files(
        self,
        get_repo_branch_mock,
        get_deployment_config_helm_chart_environment_values_file_mock,
        get_deployed_host_urls_mock,
        argocd_app_sync_mock,
        argocd_app_create_or_update_mock,
        argocd_sign_in_mock,
        git_commit_file_mock,
        clone_repo_mock,
        argocd_add_target_cluster_mock,
        git_tag_and_push_deployment_config_repo_mock,
        get_deployment_config_repo_tag_mock,
        update_yaml_file_value_mock,
        get_app_name_mock,
        argocd_get_app_manifest_mock
    ):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            step_config = {
                'argocd-username': 'argo-username',
                'argocd-password': 'argo-password',
                'argocd-api': 'https://argo.ploigos.xyz',
                'argocd-skip-tls': False,
                'deployment-config-repo': 'https://git.ploigos.xyz/foo/deploy-config',
                'deployment-config-helm-chart-path': 'charts/foo',
                'deployment-config-helm-chart-values-file-image-tag-yq-path': 'image.tag',
                'git-email': 'git@ploigos.xyz',
                'git-name': 'Ploigos',
                'container-image-tag': 'v0.42.0',
                'additional-helm-values-files': ['secrets.yaml', 'extra-secrets.yaml']
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
                environment='PROD'
            )

            actual_step_results = step_implementer._run_step()
            expected_step_result = StepResult(
                step_name='deploy',
                sub_step_name='ArgoCD',
                sub_step_implementer_name='ArgoCD',
                environment='PROD'
            )
            expected_step_result.add_artifact(
                name='argocd-app-name',
                value='test-app-name'
            )
            expected_step_result.add_artifact(
                name='config-repo-git-tag',
                value='v0.42.0'
            )
            expected_step_result.add_artifact(
                name='argocd-deployed-manifest',
                value='/does/not/matter/manifest.yaml'
            )
            expected_step_result.add_artifact(
                name='deployed-host-urls',
                value=['https://fruits.ploigos.xyz']
            )
            self.assertEqual(actual_step_results, expected_step_result)

            get_repo_branch_mock.assert_called_once_with()
            get_deployment_config_helm_chart_environment_values_file_mock.assert_called_once_with()
            get_app_name_mock.assert_called_once_with()
            deployment_config_repo_dir = os.path.join(
                step_implementer.work_dir_path,
                'deployment-config-repo'
            )
            clone_repo_mock.assert_called_once_with(
                repo_dir=deployment_config_repo_dir,
                repo_url=step_config['deployment-config-repo'],
                repo_branch='feature/test',
                git_email=step_config['git-email'],
                git_name=step_config['git-name'],
                username=None,
                password=None
            )
            update_yaml_file_value_mock.assert_called_once_with(
                file='/does/not/matter/charts/foo/values-PROD.yaml',
                yq_path=step_config['deployment-config-helm-chart-values-file-image-tag-yq-path'],
                value=step_config['container-image-tag']
            )
            git_commit_file_mock.assert_called_once_with(
                git_commit_message='Updating values for deployment to PROD',
                file_path='charts/foo/values-PROD.yaml',
                repo_dir='/does/not/matter'
            )
            get_deployment_config_repo_tag_mock.assert_called_once_with()
            git_tag_and_push_deployment_config_repo_mock.assert_called_once_with(
                deployment_config_repo=step_config['deployment-config-repo'],
                deployment_config_repo_dir='/does/not/matter',
                deployment_config_repo_tag='v0.42.0',
                force_push_tags=False
            )
            argocd_sign_in_mock.assert_called_once_with(
                argocd_api=step_config['argocd-api'],
                username=step_config['argocd-username'],
                password=step_config['argocd-password'],
                insecure=step_config['argocd-skip-tls']
            )
            argocd_add_target_cluster_mock.assert_called_once_with(
                kube_api='https://kubernetes.default.svc',
                kube_api_token=None,
                kube_api_skip_tls=False
            )
            argocd_app_create_or_update_mock.assert_called_once_with(
                argocd_app_name='test-app-name',
                repo=step_config['deployment-config-repo'],
                revision='feature/test',
                path=step_config['deployment-config-helm-chart-path'],
                dest_server='https://kubernetes.default.svc',
                auto_sync=True,
                values_files=['values-PROD.yaml', 'secrets.yaml', 'extra-secrets.yaml']
            )
            argocd_app_sync_mock.assert_called_once_with(
                argocd_app_name='test-app-name',
                argocd_sync_timeout_seconds=60,
                argocd_sync_retry_limit=3
            )
            argocd_get_app_manifest_mock.assert_called_once_with(
                argocd_app_name='test-app-name'
            )
            get_deployed_host_urls_mock.assert_called_once_with(
                manifest_path='/does/not/matter/manifest.yaml'
            )

    @patch.object(
        ArgoCD,
        '_ArgoCD__argocd_get_app_manifest',
        return_value='/does/not/matter/manifest.yaml'
    )
    @patch.object(ArgoCD, '_ArgoCD__get_app_name', return_value='test-app-name')
    @patch.object(ArgoCD, '_ArgoCD__update_yaml_file_value')
    @patch.object(ArgoCD, '_ArgoCD__get_deployment_config_repo_tag', return_value='v0.42.0')
    @patch.object(ArgoCD, '_ArgoCD__git_tag_and_push_deployment_config_repo')
    @patch.object(ArgoCD, '_ArgoCD__argocd_add_target_cluster')
    @patch.object(ArgoCD, '_ArgoCD__clone_repo', return_value='/does/not/matter')
    @patch.object(ArgoCD, '_ArgoCD__git_commit_file')
    @patch.object(ArgoCD, '_ArgoCD__argocd_sign_in')
    @patch.object(ArgoCD, '_ArgoCD__argocd_app_create_or_update')
    @patch.object(ArgoCD, '_ArgoCD__argocd_app_sync')
    @patch.object(
        ArgoCD,
        '_ArgoCD__get_deployed_host_urls',
        return_value=['https://fruits.ploigos.xyz']
    )
    @patch.object(
        ArgoCD,
        '_ArgoCD__get_deployment_config_helm_chart_environment_values_file',
        return_value='values-PROD.yaml'
    )
    @patch.object(ArgoCD, '_ArgoCD__get_repo_branch', return_value='feature/test')
    def test_ArgoCD_run_step_fail_clone_config_repo(
        self,
        get_repo_branch_mock,
        get_deployment_config_helm_chart_environment_values_file_mock,
        get_deployed_host_urls_mock,
        argocd_app_sync_mock,
        argocd_app_create_or_update_mock,
        argocd_sign_in_mock,
        git_commit_file_mock,
        clone_repo_mock,
        argocd_add_target_cluster_mock,
        git_tag_and_push_deployment_config_repo_mock,
        get_deployment_config_repo_tag_mock,
        update_yaml_file_value_mock,
        get_app_name_mock,
        argocd_get_app_manifest_mock
    ):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            step_config = {
                'argocd-username': 'argo-username',
                'argocd-password': 'argo-password',
                'argocd-api': 'https://argo.ploigos.xyz',
                'argocd-skip-tls': False,
                'deployment-config-repo': 'https://git.ploigos.xyz/foo/deploy-config',
                'deployment-config-helm-chart-path': 'charts/foo',
                'deployment-config-helm-chart-values-file-image-tag-yq-path': 'image.tag',
                'git-email': 'git@ploigos.xyz',
                'git-name': 'Ploigos',
                'container-image-tag': 'v0.42.0'
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
                environment='PROD'
            )

            clone_repo_mock.side_effect = StepRunnerException('mock failed to clone repo error')

            actual_step_results = step_implementer._run_step()
            expected_step_result = StepResult(
                step_name='deploy',
                sub_step_name='ArgoCD',
                sub_step_implementer_name='ArgoCD',
                environment='PROD'
            )
            expected_step_result.success = False
            expected_step_result.message = 'Error deploying to environment (PROD):' \
                ' mock failed to clone repo error'
            expected_step_result.add_artifact(
                name='argocd-app-name',
                value='test-app-name'
            )

            self.assertEqual(actual_step_results, expected_step_result)

            get_repo_branch_mock.assert_called_once_with()
            get_deployment_config_helm_chart_environment_values_file_mock.assert_called_once_with()
            get_app_name_mock.assert_called_once_with()
            deployment_config_repo_dir = os.path.join(
                step_implementer.work_dir_path,
                'deployment-config-repo'
            )
            clone_repo_mock.assert_called_once_with(
                repo_dir=deployment_config_repo_dir,
                repo_url=step_config['deployment-config-repo'],
                repo_branch='feature/test',
                git_email=step_config['git-email'],
                git_name=step_config['git-name'],
                username=None,
                password=None
            )
            update_yaml_file_value_mock.assert_not_called()
            git_commit_file_mock.assert_not_called()
            get_deployment_config_repo_tag_mock.assert_not_called()
            git_tag_and_push_deployment_config_repo_mock.assert_not_called()
            argocd_sign_in_mock.assert_not_called()
            argocd_add_target_cluster_mock.assert_not_called()
            argocd_app_create_or_update_mock.assert_not_called()
            argocd_app_sync_mock.assert_not_called()
            argocd_get_app_manifest_mock.assert_not_called()
            get_deployed_host_urls_mock.assert_not_called()

class TestStepImplementerDeployArgoCD__get_deployment_config_helm_chart_environment_values_file(
    TestStepImplementerDeployArgoCDBase
):
    def test_ArgoCD__get_deployment_config_helm_chart_environment_values_file_config_value(self):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            step_config = {
                'deployment-config-helm-chart-environment-values-file': 'values-AWESOME.yaml'
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )

            deployment_config_helm_chart_env_value_file = \
                step_implementer._ArgoCD__get_deployment_config_helm_chart_environment_values_file()
            self.assertEqual(
                deployment_config_helm_chart_env_value_file,
                'values-AWESOME.yaml'
            )

    def test_ArgoCD__get_deployment_config_helm_chart_environment_values_file_no_config_value_no_env(self):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            step_config = {
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path
            )

            deployment_config_helm_chart_env_value_file = \
                step_implementer._ArgoCD__get_deployment_config_helm_chart_environment_values_file()
            self.assertEqual(
                deployment_config_helm_chart_env_value_file,
                'values.yaml'
            )

    def test_ArgoCD__get_deployment_config_helm_chart_environment_values_file_no_config_value_with_env(self):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            step_config = {
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
                environment='PROD'
            )

            deployment_config_helm_chart_env_value_file = \
                step_implementer._ArgoCD__get_deployment_config_helm_chart_environment_values_file()
            self.assertEqual(
                deployment_config_helm_chart_env_value_file,
                'values-PROD.yaml'
            )

class TestStepImplementerDeployArgoCD__update_yaml_file_value(TestStepImplementerDeployArgoCDBase):
    @patch('sh.yq', create=True)
    def test_ArgoCD__update_yaml_file_value_success(self, yq_mock):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            step_config = {
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )

            updated_file_path = step_implementer._ArgoCD__update_yaml_file_value(
                file='/does/not/matter/chart/values-PROD.yaml',
                yq_path='image.tag',
                value='v0.42.0-abc123'
            )
            self.assertEqual(updated_file_path, '/does/not/matter/chart/values-PROD.yaml')
            yq_script_file_path = os.path.join(
                step_implementer.work_dir_path,
                'update-yaml-file-yq-script.yaml'
            )
            yq_mock.write.assert_called_once_with(
                '/does/not/matter/chart/values-PROD.yaml',
                f'--script={yq_script_file_path}',
                '--inplace'
            )
            with open(yq_script_file_path, 'r') as yq_script_file:
                actual_yq_script = yq_script_file.read()

                self.assertEqual(
                    actual_yq_script,
                    f"- command: update\n"
                    f"  path: image.tag\n"
                    f"  value:\n"
                    f"    v0.42.0-abc123 # written by ploigos-step-runner\n",
                )

    @patch('sh.yq', create=True)
    def test_ArgoCD__update_yaml_file_value_fail(self, yq_mock):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            step_config = {
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )

            yq_mock.write.side_effect = create_sh_side_effect(
                exception=sh.ErrorReturnCode('yq', b'mock out', b'mock yq write error')
            )

            file = '/does/not/matter/chart/values-PROD.yaml'
            yq_path = 'image.tag'
            value = 'v0.42.0-abc123'
            with self.assertRaisesRegex(
                StepRunnerException,
                re.compile(
                    rf"Error updating YAML file \({file}\) target \({yq_path}\)"
                    rf" with value \({value}\):"
                    r".*RAN: yq"
                    r".*STDOUT:"
                    r".*mock out"
                    r".*STDERR:"
                    r".*mock yq write error",
                    re.DOTALL
                )
            ):
                step_implementer._ArgoCD__update_yaml_file_value(
                    file=file,
                    yq_path=yq_path,
                    value=value
                )
                yq_script_file_path = os.path.join(
                    step_implementer.work_dir_path,
                    'update-yaml-file-yq-script.yaml'
                )
                yq_mock.write.assert_called_once_with(
                    '/does/not/matter/chart/values-PROD.yaml',
                    f'--script={yq_script_file_path}',
                    '--inplace'
                )
                with open(yq_script_file_path, 'r') as yq_script_file:
                    actual_yq_script = yq_script_file.read()

                    self.assertEqual(
                        actual_yq_script,
                        f"- command: update\n"
                        f"  path: image.tag\n"
                        f"  value:\n"
                        f"    v0.42.0-abc123 # written by ploigos-step-runner\n",
                    )

class TestStepImplementerDeployArgoCD__git_tag_and_push_deployment_config_repo(TestStepImplementerDeployArgoCDBase):
    @patch.object(ArgoCD, '_ArgoCD__git_tag_and_push')
    def test_ArgoCD__git_tag_and_push_deployment_config_repo_http(self, git_tag_and_push_mock):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            step_config = {
                'git-username': 'test-username',
                'git-password': 'test-password'
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )

            step_implementer._ArgoCD__git_tag_and_push_deployment_config_repo(
                deployment_config_repo='http://git.ploigos.xys/foo',
                deployment_config_repo_dir='/does/not/matter',
                deployment_config_repo_tag='test-tag',
                force_push_tags=False
            )
            git_tag_and_push_mock.assert_called_once_with(
                repo_dir='/does/not/matter',
                tag='test-tag',
                url='http://test-username:test-password@git.ploigos.xys/foo',
                force_push_tags=False
            )

    @patch.object(ArgoCD, '_ArgoCD__git_tag_and_push')
    def test_ArgoCD__git_tag_and_push_deployment_config_repo_https(self, git_tag_and_push_mock):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            step_config = {
                'git-username': 'test-username',
                'git-password': 'test-password'
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )

            step_implementer._ArgoCD__git_tag_and_push_deployment_config_repo(
                deployment_config_repo='https://git.ploigos.xys/foo',
                deployment_config_repo_dir='/does/not/matter',
                deployment_config_repo_tag='test-tag',
                force_push_tags=False
            )
            git_tag_and_push_mock.assert_called_once_with(
                repo_dir='/does/not/matter',
                tag='test-tag',
                url='https://test-username:test-password@git.ploigos.xys/foo',
                force_push_tags=False
            )

    @patch.object(ArgoCD, '_ArgoCD__git_tag_and_push')
    def test_ArgoCD__git_tag_and_push_deployment_config_repo_ssh(self, git_tag_and_push_mock):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            step_config = {
                'git-username': 'test-username',
                'git-password': 'test-password'
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )

            step_implementer._ArgoCD__git_tag_and_push_deployment_config_repo(
                deployment_config_repo='git@git.ploigos.xys:foo/bar',
                deployment_config_repo_dir='/does/not/matter',
                deployment_config_repo_tag='test-tag',
                force_push_tags=False
            )
            git_tag_and_push_mock.assert_called_once_with(
                repo_dir='/does/not/matter',
                tag='test-tag',
                force_push_tags=False
            )


class TestStepImplementerDeployArgoCD__get_app_name(TestStepImplementerDeployArgoCDBase):
    @patch.object(ArgoCD, '_ArgoCD__get_repo_branch', return_value='feature/test')
    def test_ArgoCD__get_app_name_no_env_less_then_max_length(self, get_repo_branch_mock):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            step_config = {
                'organization': 'test-org',
                'application-name': 'test-app',
                'service-name': 'test-service'
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )

            app_name = step_implementer._ArgoCD__get_app_name()
            self.assertEqual(app_name, 'test-org-test-app-test-service-feature-test')

    @patch.object(ArgoCD, '_ArgoCD__get_repo_branch', return_value='feature/TEST')
    def test_ArgoCD__get_app_name_no_env_less_then_max_length_caps(self, get_repo_branch_mock):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            step_config = {
                'organization': 'test-ORG',
                'application-name': 'test-APP',
                'service-name': 'test-SERVICE'
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )

            app_name = step_implementer._ArgoCD__get_app_name()
            self.assertEqual(app_name, 'test-org-test-app-test-service-feature-test')

    @patch.object(ArgoCD, '_ArgoCD__get_repo_branch', return_value='feature/test')
    def test_ArgoCD__get_app_name_no_env_more_then_max_length(self, get_repo_branch_mock):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            step_config = {
                'organization': 'test-org',
                'application-name': 'test-app',
                'service-name': 'test-service-this-is-really-long-hello-world-foo'
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )

            app_name = step_implementer._ArgoCD__get_app_name()
            self.assertEqual(app_name, 'ice-this-is-really-long-hello-world-foo-feature-test')

    @patch.object(ArgoCD, '_ArgoCD__get_repo_branch', return_value='feature/test')
    def test_ArgoCD__get_app_name_with_env_less_then_max_length(self, get_repo_branch_mock):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            step_config = {
                'organization': 'test-org',
                'application-name': 'test-app',
                'service-name': 'test-service'
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
                environment='PROD'
            )

            app_name = step_implementer._ArgoCD__get_app_name()
            self.assertEqual(app_name, 'test-org-test-app-test-service-feature-test-prod')

class TestStepImplementerDeployArgoCD__get_deployment_config_repo_tag(TestStepImplementerDeployArgoCDBase):
    def test_ArgoCD__get_deployment_config_repo_tag_use_tag(self):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            step_config = {
                'tag': 'v0.42.0-abc123',
                'version': 'v0.42.0'
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )

            deployment_config_repo_tag = step_implementer._ArgoCD__get_deployment_config_repo_tag()
            self.assertEqual(deployment_config_repo_tag, 'v0.42.0-abc123')

    def test_ArgoCD__get_deployment_config_repo_tag_use_version(self):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            step_config = {
                'version': 'v0.42.0'
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )

            deployment_config_repo_tag = step_implementer._ArgoCD__get_deployment_config_repo_tag()
            self.assertEqual(deployment_config_repo_tag, 'v0.42.0')


    def test_ArgoCD__get_deployment_config_repo_tag_use_latest(self):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            step_config = {
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )

            deployment_config_repo_tag = step_implementer._ArgoCD__get_deployment_config_repo_tag()
            self.assertEqual(deployment_config_repo_tag, 'latest')

    def test_ArgoCD__get_deployment_config_repo_tag_use_tag_with_env(self):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            step_config = {
                'tag': 'v0.42.0-main+abc123',
                'version': 'v0.42.0'
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
                environment='PROD'
            )

            deployment_config_repo_tag = step_implementer._ArgoCD__get_deployment_config_repo_tag()
            self.assertEqual(deployment_config_repo_tag, 'v0.42.0-main+abc123.PROD')

class TestStepImplementerDeployArgoCD__get_deployed_host_urls(TestStepImplementerDeployArgoCDBase):
    def __run__ArgoCD__get_deployed_host_urls_test(
        self,
        manifest_contents,
        expected_host_urls
    ):
        with TempDirectory() as temp_dir:
            temp_dir.write('manifest.yaml', bytes(manifest_contents, 'utf-8'))
            manifest_path = os.path.join(temp_dir.path, 'manifest.yaml')

            actual_host_urls = ArgoCD._ArgoCD__get_deployed_host_urls(
                manifest_path=manifest_path
            )

            self.assertEqual(actual_host_urls, expected_host_urls)

    def test_ArgoCD__get_deployed_host_urls_empty_file(self):
        self.__run__ArgoCD__get_deployed_host_urls_test(
            manifest_contents="",
            expected_host_urls=[]
        )

    def test_ArgoCD__get_deployed_host_urls_empty_yaml_doc(self):
        self.__run__ArgoCD__get_deployed_host_urls_test(
            manifest_contents="---",
            expected_host_urls=[]
        )

    def test_ArgoCD__get_deployed_host_urls_yaml_doc_no_kind(self):
        self.__run__ArgoCD__get_deployed_host_urls_test(
            manifest_contents="""
---
foo: test"
""",
            expected_host_urls=[]
        )

    def test_ArgoCD__get_deployed_host_urls_1_http_routes_no_ingress(self):
        self.__run__ArgoCD__get_deployed_host_urls_test(
            manifest_contents="""
---
apiVersion: route.openshift.io/v1
kind: Route
metadata:
  annotations:
    kubectl.kubernetes.io/last-applied-configuration: |
      {"apiVersion":"route.openshift.io/v1","kind":"Route","metadata":{"annotations":{},"labels":{"app.kubernetes.io/instance":"s-mvn-jenkins-std-fruit-feature-results-refactor-dev"},"name":"fruit","namespace":"s-mvn-jenkins-std-fruit-feature-results-refactor-dev"},"spec":{"path":"/","port":{"targetPort":"http"},"to":{"kind":"Service","name":"fruit"}}}
    openshift.io/host.generated: "true"
  creationTimestamp: "2020-12-16T22:14:46Z"
  labels:
    app.kubernetes.io/instance: s-mvn-jenkins-std-fruit-feature-results-refactor-dev
  managedFields:
  - apiVersion: route.openshift.io/v1
    fieldsType: FieldsV1
    fieldsV1:
      f:metadata:
        f:annotations:
          .: {}
          f:kubectl.kubernetes.io/last-applied-configuration: {}
        f:labels:
          .: {}
          f:app.kubernetes.io/instance: {}
      f:spec:
        f:path: {}
        f:port:
          .: {}
          f:targetPort: {}
        f:to:
          f:kind: {}
          f:name: {}
          f:weight: {}
        f:wildcardPolicy: {}
    manager: argocd-application-controller
    operation: Update
    time: "2020-12-16T22:14:46Z"
  - apiVersion: route.openshift.io/v1
    fieldsType: FieldsV1
    fieldsV1:
      f:status:
        f:ingress: {}
    manager: openshift-router
    operation: Update
    time: "2020-12-16T22:14:46Z"
  name: fruit
  namespace: s-mvn-jenkins-std-fruit-feature-results-refactor-dev
  resourceVersion: "92938959"
  selfLink: /apis/route.openshift.io/v1/namespaces/s-mvn-jenkins-std-fruit-feature-results-refactor-dev/routes/fruit
  uid: f5d36e42-b894-49d9-9885-64ed7bc438db
spec:
  host: fruit-s-mvn-jenkins-std-fruit-feature-results-refactor-dev.apps.ploigos.xyz
  path: /
  port:
    targetPort: http
  to:
    kind: Service
    name: fruit
    weight: 100
  wildcardPolicy: None
status:
  ingress:
  - conditions:
    - lastTransitionTime: "2020-12-16T22:14:46Z"
      status: "True"
      type: Admitted
    host: fruit-s-mvn-jenkins-std-fruit-feature-results-refactor-dev.apps.ploigos.xyz
    routerCanonicalHostname: apps.ploigos.xyz
    routerName: default
    wildcardPolicy: None
""",
            expected_host_urls=[
                'http://fruit-s-mvn-jenkins-std-fruit-feature-results-refactor-dev.apps.ploigos.xyz'
            ]
        )

    def test_ArgoCD__get_deployed_host_urls_1_https_routes_no_ingress(self):
        self.__run__ArgoCD__get_deployed_host_urls_test(
            manifest_contents="""
---
apiVersion: route.openshift.io/v1
kind: Route
metadata:
  annotations:
    kubectl.kubernetes.io/last-applied-configuration: |
      {"apiVersion":"route.openshift.io/v1","kind":"Route","metadata":{"annotations":{},"labels":{"app.kubernetes.io/instance":"s-mvn-jenkins-std-fruit-feature-results-refactor-dev"},"name":"fruit","namespace":"s-mvn-jenkins-std-fruit-feature-results-refactor-dev"},"spec":{"path":"/","port":{"targetPort":"http"},"to":{"kind":"Service","name":"fruit"}}}
    openshift.io/host.generated: "true"
  creationTimestamp: "2020-12-16T22:14:46Z"
  labels:
    app.kubernetes.io/instance: s-mvn-jenkins-std-fruit-feature-results-refactor-dev
  managedFields:
  - apiVersion: route.openshift.io/v1
    fieldsType: FieldsV1
    fieldsV1:
      f:metadata:
        f:annotations:
          .: {}
          f:kubectl.kubernetes.io/last-applied-configuration: {}
        f:labels:
          .: {}
          f:app.kubernetes.io/instance: {}
      f:spec:
        f:path: {}
        f:port:
          .: {}
          f:targetPort: {}
        f:to:
          f:kind: {}
          f:name: {}
          f:weight: {}
        f:wildcardPolicy: {}
    manager: argocd-application-controller
    operation: Update
    time: "2020-12-16T22:14:46Z"
  - apiVersion: route.openshift.io/v1
    fieldsType: FieldsV1
    fieldsV1:
      f:status:
        f:ingress: {}
    manager: openshift-router
    operation: Update
    time: "2020-12-16T22:14:46Z"
  name: fruit
  namespace: s-mvn-jenkins-std-fruit-feature-results-refactor-dev
  resourceVersion: "92938959"
  selfLink: /apis/route.openshift.io/v1/namespaces/s-mvn-jenkins-std-fruit-feature-results-refactor-dev/routes/fruit
  uid: f5d36e42-b894-49d9-9885-64ed7bc438db
spec:
  host: fruit-s-mvn-jenkins-std-fruit-feature-results-refactor-dev.apps.ploigos.xyz
  path: /
  port:
    targetPort: http
  to:
    kind: Service
    name: fruit
    weight: 100
  tls:
    termination: edge
  wildcardPolicy: None
status:
  ingress:
  - conditions:
    - lastTransitionTime: "2020-12-16T22:14:46Z"
      status: "True"
      type: Admitted
    host: fruit-s-mvn-jenkins-std-fruit-feature-results-refactor-dev.apps.ploigos.xyz
    routerCanonicalHostname: apps.ploigos.xyz
    routerName: default
    wildcardPolicy: None
""",
            expected_host_urls=[
                'https://fruit-s-mvn-jenkins-std-fruit-feature-results-refactor-dev.apps.ploigos.xyz'
            ]
        )

    def test_ArgoCD__get_deployed_host_urls_no_routes_1_http_ingress(self):
        self.__run__ArgoCD__get_deployed_host_urls_test(
            manifest_contents="""
---
kind: Ingress
apiVersion: networking.k8s.io/v1
metadata:
  name: test
spec:
  rules:
  - host: foo.apps.ploigos.xyz
    http:
      paths:
      - path: /
        pathType: Prefix
        backend:
          service:
            name: foo
            port:
              name: http-listener
""",
            expected_host_urls=[
                'http://foo.apps.ploigos.xyz'
            ]
        )

    def test_ArgoCD__get_deployed_host_urls_no_routes_1_http_ingress_with_tls_list(self):
        self.__run__ArgoCD__get_deployed_host_urls_test(
            manifest_contents="""
---
kind: Ingress
apiVersion: networking.k8s.io/v1
metadata:
  name: test
spec:
  tls:
  - hosts:
    - does-not-match.apps.ploigos.xyz
  rules:
  - host: foo.apps.ploigos.xyz
    http:
      paths:
      - path: /
        pathType: Prefix
        backend:
          service:
            name: foo
            port:
              name: http-listener
""",
            expected_host_urls=[
                'http://foo.apps.ploigos.xyz'
            ]
        )

    def test_ArgoCD__get_deployed_host_urls_no_routes_1_https_ingress(self):
        self.__run__ArgoCD__get_deployed_host_urls_test(
            manifest_contents="""
---
kind: Ingress
apiVersion: networking.k8s.io/v1
metadata:
  name: test
  annotations:
    route.openshift.io/termination: "edge"
spec:
  tls:
  - hosts:
    - foo.apps.ploigos.xyz
  rules:
  - host: foo.apps.ploigos.xyz
    http:
      paths:
      - path: /
        pathType: Prefix
        backend:
          service:
            name: foo
            port:
              name: http-listener
""",
            expected_host_urls=[
                'https://foo.apps.ploigos.xyz'
            ]
        )

    def test_ArgoCD__get_deployed_host_urls_1_http_route_1_https_route_1_http_ingress_1_https_ingress(self):
        self.__run__ArgoCD__get_deployed_host_urls_test(
            manifest_contents="""
---
kind: Ingress
apiVersion: networking.k8s.io/v1
metadata:
  name: test
spec:
  rules:
  - host: ingress1.apps.ploigos.xyz
    http:
      paths:
      - path: /
        pathType: Prefix
        backend:
          service:
            name: foo
            port:
              name: http-listener
---
kind: Ingress
apiVersion: networking.k8s.io/v1
metadata:
  name: test
  annotations:
    route.openshift.io/termination: "edge"
spec:
  tls:
  - hosts:
    - ingress2.apps.ploigos.xyz
  rules:
  - host: ingress2.apps.ploigos.xyz
    http:
      paths:
      - path: /
        pathType: Prefix
        backend:
          service:
            name: foo
            port:
              name: http-listener
---
apiVersion: route.openshift.io/v1
kind: Route
metadata:
  annotations:
    kubectl.kubernetes.io/last-applied-configuration: |
      {"apiVersion":"route.openshift.io/v1","kind":"Route","metadata":{"annotations":{},"labels":{"app.kubernetes.io/instance":"s-mvn-jenkins-std-fruit-feature-results-refactor-dev"},"name":"fruit","namespace":"s-mvn-jenkins-std-fruit-feature-results-refactor-dev"},"spec":{"path":"/","port":{"targetPort":"http"},"to":{"kind":"Service","name":"fruit"}}}
    openshift.io/host.generated: "true"
  creationTimestamp: "2020-12-16T22:14:46Z"
  labels:
    app.kubernetes.io/instance: s-mvn-jenkins-std-fruit-feature-results-refactor-dev
  managedFields:
  - apiVersion: route.openshift.io/v1
    fieldsType: FieldsV1
    fieldsV1:
      f:metadata:
        f:annotations:
          .: {}
          f:kubectl.kubernetes.io/last-applied-configuration: {}
        f:labels:
          .: {}
          f:app.kubernetes.io/instance: {}
      f:spec:
        f:path: {}
        f:port:
          .: {}
          f:targetPort: {}
        f:to:
          f:kind: {}
          f:name: {}
          f:weight: {}
        f:wildcardPolicy: {}
    manager: argocd-application-controller
    operation: Update
    time: "2020-12-16T22:14:46Z"
  - apiVersion: route.openshift.io/v1
    fieldsType: FieldsV1
    fieldsV1:
      f:status:
        f:ingress: {}
    manager: openshift-router
    operation: Update
    time: "2020-12-16T22:14:46Z"
  name: fruit
  namespace: s-mvn-jenkins-std-fruit-feature-results-refactor-dev
  resourceVersion: "92938959"
  selfLink: /apis/route.openshift.io/v1/namespaces/s-mvn-jenkins-std-fruit-feature-results-refactor-dev/routes/fruit
  uid: f5d36e42-b894-49d9-9885-64ed7bc438db
spec:
  host: route1.apps.ploigos.xyz
  path: /
  port:
    targetPort: http
  to:
    kind: Service
    name: fruit
    weight: 100
  tls:
    termination: edge
  wildcardPolicy: None
status:
  ingress:
  - conditions:
    - lastTransitionTime: "2020-12-16T22:14:46Z"
      status: "True"
      type: Admitted
    host: route1.apps.ploigos.xyz
    routerCanonicalHostname: apps.ploigos.xyz
    routerName: default
    wildcardPolicy: None
---
apiVersion: route.openshift.io/v1
kind: Route
metadata:
  annotations:
    kubectl.kubernetes.io/last-applied-configuration: |
      {"apiVersion":"route.openshift.io/v1","kind":"Route","metadata":{"annotations":{},"labels":{"app.kubernetes.io/instance":"s-mvn-jenkins-std-fruit-feature-results-refactor-dev"},"name":"fruit","namespace":"s-mvn-jenkins-std-fruit-feature-results-refactor-dev"},"spec":{"path":"/","port":{"targetPort":"http"},"to":{"kind":"Service","name":"fruit"}}}
    openshift.io/host.generated: "true"
  creationTimestamp: "2020-12-16T22:14:46Z"
  labels:
    app.kubernetes.io/instance: s-mvn-jenkins-std-fruit-feature-results-refactor-dev
  managedFields:
  - apiVersion: route.openshift.io/v1
    fieldsType: FieldsV1
    fieldsV1:
      f:metadata:
        f:annotations:
          .: {}
          f:kubectl.kubernetes.io/last-applied-configuration: {}
        f:labels:
          .: {}
          f:app.kubernetes.io/instance: {}
      f:spec:
        f:path: {}
        f:port:
          .: {}
          f:targetPort: {}
        f:to:
          f:kind: {}
          f:name: {}
          f:weight: {}
        f:wildcardPolicy: {}
    manager: argocd-application-controller
    operation: Update
    time: "2020-12-16T22:14:46Z"
  - apiVersion: route.openshift.io/v1
    fieldsType: FieldsV1
    fieldsV1:
      f:status:
        f:ingress: {}
    manager: openshift-router
    operation: Update
    time: "2020-12-16T22:14:46Z"
  name: fruit
  namespace: s-mvn-jenkins-std-fruit-feature-results-refactor-dev
  resourceVersion: "92938959"
  selfLink: /apis/route.openshift.io/v1/namespaces/s-mvn-jenkins-std-fruit-feature-results-refactor-dev/routes/fruit
  uid: f5d36e42-b894-49d9-9885-64ed7bc438db
spec:
  host: route2.apps.ploigos.xyz
  path: /
  port:
    targetPort: http
  to:
    kind: Service
    name: fruit
    weight: 100
  wildcardPolicy: None
status:
  ingress:
  - conditions:
    - lastTransitionTime: "2020-12-16T22:14:46Z"
      status: "True"
      type: Admitted
    host: route2.apps.ploigos.xyz.apps.ploigos.xyz
    routerCanonicalHostname: apps.ploigos.xyz
    routerName: default
    wildcardPolicy: None
""",
            expected_host_urls=[
                'http://ingress1.apps.ploigos.xyz',
                'https://ingress2.apps.ploigos.xyz',
                'https://route1.apps.ploigos.xyz',
                'http://route2.apps.ploigos.xyz'
            ]
        )

class TestStepImplementerDeployArgoCD__clone_repo(TestStepImplementerDeployArgoCDBase):
    @patch.object(sh, 'git')
    def test_ArgoCD__clone_repo_success_new_branch(self, git_mock):
        repo_dir = '/does/not/matter'
        repo_url = 'git@git.ploigos.xyz:/foo/test.git'
        repo_branch = 'feature/test'
        git_email = 'test@ploigos.xyz'
        git_name = 'Test Robot'
        username = 'Test'
        password = 'Password'
        ArgoCD._ArgoCD__clone_repo(
            repo_dir=repo_dir,
            repo_url=repo_url,
            repo_branch=repo_branch,
            git_email=git_email,
            git_name=git_name,
            username=username,
            password=password
        )

        git_mock.clone.assert_called_once_with(
            repo_url,
            repo_dir,
            _out=Any(IOBase),
            _err=Any(IOBase)
        )
        git_mock.checkout.assert_called_once_with(
            repo_branch,
            _cwd=repo_dir,
            _out=Any(IOBase),
            _err=Any(IOBase)
        )
        git_mock.config.assert_has_calls([
            call(
                'user.email',
                git_email,
                _cwd=repo_dir,
                _out=Any(IOBase),
                _err=Any(IOBase)
            ),
            call(
                'user.name',
                git_name,
                _cwd=repo_dir,
                _out=Any(IOBase),
                _err=Any(IOBase)
            )
        ])

    @patch.object(sh, 'git')
    def test_ArgoCD__clone_repo_success_new_branch_https(self, git_mock):
        repo_dir = '/does/not/matter'
        repo_url = 'https://Test:Password@git.ploigos.xyz'
        repo_branch = 'feature/test'
        git_email = 'test@ploigos.xyz'
        git_name = 'Test Robot'
        username = 'Test'
        password = 'Password'
        repo_url_with_auth = 'https://git.ploigos.xyz'
        ArgoCD._ArgoCD__clone_repo(
            repo_dir=repo_dir,
            repo_url=repo_url_with_auth,
            repo_branch=repo_branch,
            git_email=git_email,
            git_name=git_name,
            username=username,
            password=password
        )

        git_mock.clone.assert_called_once_with(
            repo_url,
            repo_dir,
            _out=Any(IOBase),
            _err=Any(IOBase)
        )
        git_mock.checkout.assert_called_once_with(
            repo_branch,
            _cwd=repo_dir,
            _out=Any(IOBase),
            _err=Any(IOBase)
        )
        git_mock.config.assert_has_calls([
            call(
                'user.email',
                git_email,
                _cwd=repo_dir,
                _out=Any(IOBase),
                _err=Any(IOBase)
            ),
            call(
                'user.name',
                git_name,
                _cwd=repo_dir,
                _out=Any(IOBase),
                _err=Any(IOBase)
            )
        ])


    @patch.object(sh, 'git')
    def test_ArgoCD__clone_repo_success_existing_branch(self, git_mock):
        repo_dir = '/does/not/matter'
        repo_url = 'git@git.ploigos.xyz:/foo/test.git'
        repo_branch = 'feature/test'
        git_email = 'test@ploigos.xyz'
        git_name = 'Test Robot'
        username = 'Test'
        password = 'password'

        git_mock.checkout.side_effect = [
            sh.ErrorReturnCode('git', b'mock out', b'mock git checkout branch does not exist'),
            None
        ]

        ArgoCD._ArgoCD__clone_repo(
            repo_dir=repo_dir,
            repo_url=repo_url,
            repo_branch=repo_branch,
            git_email=git_email,
            git_name=git_name,
            username=username,
            password=password
        )

        git_mock.clone.assert_called_once_with(
            repo_url,
            repo_dir,
            _out=Any(IOBase),
            _err=Any(IOBase)
        )
        git_mock.checkout.assert_has_calls([
            call(
                repo_branch,
                _cwd=repo_dir,
                _out=Any(IOBase),
                _err=Any(IOBase)
            ),
            call(
                '-b',
                repo_branch,
                _cwd=repo_dir,
                _out=Any(IOBase),
                _err=Any(IOBase)
            )
        ])
        git_mock.config.assert_has_calls([
            call(
                'user.email',
                git_email,
                _cwd=repo_dir,
                _out=Any(IOBase),
                _err=Any(IOBase)
            ),
            call(
                'user.name',
                git_name,
                _cwd=repo_dir,
                _out=Any(IOBase),
                _err=Any(IOBase)
            )
        ])

    @patch.object(sh, 'git')
    def test_ArgoCD__clone_repo_fail_clone(self, git_mock):
        repo_dir = '/does/not/matter'
        repo_url = 'git@git.ploigos.xyz:/foo/test.git'
        repo_branch = 'feature/test'
        git_email = 'test@ploigos.xyz'
        git_name = 'Test Robot'
        username = 'Test'
        password = 'password'

        git_mock.clone.side_effect = create_sh_side_effect(
            exception=sh.ErrorReturnCode('git', b'mock out', b'mock git clone error')
        )

        with self.assertRaisesRegex(
            StepRunnerException,
            re.compile(
                rf"Error cloning repository \({repo_url}\):"
                r".*RAN: git"
                r".*STDOUT:"
                r".*mock out"
                r".*STDERR:"
                r".*mock git clone error",
                re.DOTALL
            )
        ):
            ArgoCD._ArgoCD__clone_repo(
                repo_dir=repo_dir,
                repo_url=repo_url,
                repo_branch=repo_branch,
                git_email=git_email,
                git_name=git_name,
                username=username,
                password=password
            )

            git_mock.clone.assert_called_once_with(
                repo_url,
                repo_dir,
                _out=Any(IOBase),
                _err=Any(IOBase)
            )
            git_mock.checkout.assert_not_called()
            git_mock.config.assert_not_called()

    @patch.object(sh, 'git')
    def test_ArgoCD__clone_repo_fail_existing_branch(self, git_mock):
        repo_dir = '/does/not/matter'
        repo_url = 'git@git.ploigos.xyz:/foo/test.git'
        repo_branch = 'feature/test'
        git_email = 'test@ploigos.xyz'
        git_name = 'Test Robot'
        username = 'Test'
        password = 'password'

        git_mock.checkout.side_effect = [
            sh.ErrorReturnCode('git', b'mock out', b'mock git checkout branch does not exist'),
            sh.ErrorReturnCode('git', b'mock out', b'mock git checkout new branch error'),
        ]

        with self.assertRaisesRegex(
            StepRunnerException,
            re.compile(
                rf"Unexpected error checking out new or existing branch \({repo_branch}\)"
                rf" from repository \({repo_url}\):"
                r".*RAN: git"
                r".*STDOUT:"
                r".*mock out"
                r".*STDERR:"
                r".*mock git checkout new branch error",
                re.DOTALL
            )
        ):
            ArgoCD._ArgoCD__clone_repo(
                repo_dir=repo_dir,
                repo_url=repo_url,
                repo_branch=repo_branch,
                git_email=git_email,
                git_name=git_name,
                username=username,
                password=password
            )

            git_mock.clone.assert_called_once_with(
                repo_url,
                repo_dir,
                _out=Any(IOBase),
                _err=Any(IOBase)
            )
            git_mock.checkout.assert_has_calls([
                call(
                    repo_branch,
                    _cwd=repo_dir,
                    _out=Any(IOBase),
                    _err=Any(IOBase)
                ),
                call(
                    '-b',
                    repo_branch,
                    _cwd=repo_dir,
                    _out=Any(IOBase),
                    _err=Any(IOBase)
                )
            ])
            git_mock.config.assert_not_called()

    @patch.object(sh, 'git')
    def test_ArgoCD__clone_repo_fail_config_email(self, git_mock):
        repo_dir = '/does/not/matter'
        repo_url = 'git@git.ploigos.xyz:/foo/test.git'
        repo_branch = 'feature/test'
        git_email = 'test@ploigos.xyz'
        git_name = 'Test Robot'
        username = 'Test'
        password = 'password'

        git_mock.config.side_effect = [
            sh.ErrorReturnCode('git', b'mock out', b'mock git config email error'),
            None
        ]

        with self.assertRaisesRegex(
            StepRunnerException,
            re.compile(
                rf"Unexpected error configuring git user.email \({git_email}\)"
                rf" and user.name \({git_name}\) for repository \({repo_url}\)"
                rf" in directory \({repo_dir}\):"
                r".*RAN: git"
                r".*STDOUT:"
                r".*mock out"
                r".*STDERR:"
                r".*mock git config email error",
                re.DOTALL
            )
        ):
            ArgoCD._ArgoCD__clone_repo(
                repo_dir=repo_dir,
                repo_url=repo_url,
                repo_branch=repo_branch,
                git_email=git_email,
                git_name=git_name,
                username=username,
                password=password
            )

            git_mock.clone.assert_called_once_with(
                repo_url,
                repo_dir,
                _out=Any(IOBase),
                _err=Any(IOBase)
            )
            git_mock.checkout.assert_has_calls([
                call(
                    repo_branch,
                    _cwd=repo_dir,
                    _out=Any(IOBase),
                    _err=Any(IOBase)
                ),
                call(
                    '-b',
                    repo_branch,
                    _cwd=repo_dir,
                    _out=Any(IOBase),
                    _err=Any(IOBase)
                )
            ])
            git_mock.config.assert_called_once_with(
                'user.email',
                git_email,
                _cwd=repo_dir,
                _out=Any(IOBase),
                _err=Any(IOBase)
            )

    @patch.object(sh, 'git')
    def test_ArgoCD__clone_repo_fail_config_name(self, git_mock):
        repo_dir = '/does/not/matter'
        repo_url = 'git@git.ploigos.xyz:/foo/test.git'
        repo_branch = 'feature/test'
        git_email = 'test@ploigos.xyz'
        git_name = 'Test Robot'
        username = 'Test'
        password = 'password'

        git_mock.config.side_effect = [
            None,
            sh.ErrorReturnCode('git', b'mock out', b'mock git config name error')
        ]

        with self.assertRaisesRegex(
            StepRunnerException,
            re.compile(
                rf"Unexpected error configuring git user.email \({git_email}\)"
                rf" and user.name \({git_name}\) for repository \({repo_url}\)"
                rf" in directory \({repo_dir}\):"
                r".*RAN: git"
                r".*STDOUT:"
                r".*mock out"
                r".*STDERR:"
                r".*mock git config name error",
                re.DOTALL
            )
        ):
            ArgoCD._ArgoCD__clone_repo(
                repo_dir=repo_dir,
                repo_url=repo_url,
                repo_branch=repo_branch,
                git_email=git_email,
                git_name=git_name,
                username=username,
                password=password
            )

            git_mock.clone.assert_called_once_with(
                repo_url,
                repo_dir,
                _out=Any(IOBase),
                _err=Any(IOBase)
            )
            git_mock.checkout.assert_has_calls([
                call(
                    repo_branch,
                    _cwd=repo_dir,
                    _out=Any(IOBase),
                    _err=Any(IOBase)
                ),
                call(
                    '-b',
                    repo_branch,
                    _cwd=repo_dir,
                    _out=Any(IOBase),
                    _err=Any(IOBase)
                )
            ])
            git_mock.config.assert_has_calls([
                call(
                    'user.email',
                    git_email,
                    _cwd=repo_dir,
                    _out=Any(IOBase),
                    _err=Any(IOBase)
                ),
                call(
                    'user.name',
                    git_name,
                    _cwd=repo_dir,
                    _out=Any(IOBase),
                    _err=Any(IOBase)
                )
            ])

class TestStepImplementerDeployArgoCD__get_repo_branch(TestStepImplementerDeployArgoCDBase):
    @patch.object(sh, 'git')
    def test_ArgoCD__get_repo_branch_success(self, git_mock):
        git_mock.side_effect = create_sh_side_effect(mock_stdout="feature/test")

        repo_branch = ArgoCD._ArgoCD__get_repo_branch()
        self.assertEqual(repo_branch, 'feature/test')
        git_mock.assert_called_once_with(
            'rev-parse',
            '--abbrev-ref',
            'HEAD'
        )

    @patch.object(sh, 'git')
    def test_ArgoCD__get_repo_branch_fail(self, git_mock):
        git_mock.side_effect = create_sh_side_effect(
            exception=sh.ErrorReturnCode('git', b'mock out', b'mock git rev-parse error')
        )

        with self.assertRaisesRegex(
            StepRunnerException,
            re.compile(
                r"Unexpected error getting checkedout git repository branch"
                r" of the working directory."
                r".*RAN: git"
                r".*STDOUT:"
                r".*mock out"
                r".*STDERR:"
                r".*mock git rev-parse error",
                re.DOTALL
            )
        ):
            ArgoCD._ArgoCD__get_repo_branch()
            git_mock.assert_called_once_with(
                'rev-parse',
                '--abbrev-ref',
                'HEAD'
            )

class TestStepImplementerDeployArgoCD__git_tag_and_push(TestStepImplementerDeployArgoCDBase):
    @patch.object(sh, 'git')
    def test_ArgoCD__git_tag_and_push_success_ssh(self, git_mock):
        repo_dir = '/does/not/matter'
        tag = 'v0.42.0'
        url = None
        ArgoCD._ArgoCD__git_tag_and_push(
            repo_dir=repo_dir,
            tag=tag,
            url=url
        )

        git_mock.push.assert_has_calls([
            call(
                _cwd=repo_dir,
                _out=Any(IOBase)
            ),
            call(
                '--tag',
                _cwd=repo_dir,
                _out=Any(IOBase)
            )
        ])
        git_mock.tag.assert_called_once_with(
            tag,
            '-f',
            _cwd=repo_dir,
            _out=Any(IOBase),
            _err=Any(IOBase)
        )

    @patch.object(sh, 'git')
    def test_ArgoCD__git_tag_and_push_success_https_url(self, git_mock):
        repo_dir = '/does/not/matter'
        tag = 'v0.42.0'
        url = 'https://user:pass@git.ploigos.xyz'
        ArgoCD._ArgoCD__git_tag_and_push(
            repo_dir=repo_dir,
            tag=tag,
            url=url
        )

        git_mock.push.bake().assert_has_calls([
            call(
                _cwd=repo_dir,
                _out=Any(IOBase)
            ),
            call(
                '--tag',
                _cwd=repo_dir,
                _out=Any(IOBase)
            )
        ])
        git_mock.tag.assert_called_once_with(
            tag,
            '-f',
            _cwd=repo_dir,
            _out=Any(IOBase),
            _err=Any(IOBase)
        )

    @patch.object(sh, 'git')
    def test_ArgoCD__git_tag_and_push_fail_commit(self, git_mock):
        repo_dir = '/does/not/matter'
        tag = 'v0.42.0'
        url = None

        git_mock.push.side_effect = [
            sh.ErrorReturnCode('git', b'mock out', b'mock git push error'),
            create_sh_side_effect()
        ]

        with self.assertRaisesRegex(
            StepRunnerException,
            re.compile(
                rf"Error pushing commits from repository directory \({repo_dir}\) to"
                rf" repository \({url}\):"
                r".*RAN: git"
                r".*STDOUT:"
                r".*mock out"
                r".*STDERR:"
                r".*mock git push error",
                re.DOTALL
            )
        ):
            ArgoCD._ArgoCD__git_tag_and_push(
                repo_dir=repo_dir,
                tag=tag,
                url=url
            )

            git_mock.push.assert_has_calls([
                call(
                    _cwd=repo_dir,
                    _out=Any(IOBase)
                )
            ])

            git_mock.tag.assert_not_called()

    @patch.object(sh, 'git')
    def test_ArgoCD__git_tag_and_push_fail_tag(self, git_mock):
        repo_dir = '/does/not/matter'
        tag = 'v0.42.0'
        url = None

        git_mock.tag.side_effect = sh.ErrorReturnCode('git', b'mock out', b'mock git tag error')

        with self.assertRaisesRegex(
            StepRunnerException,
            re.compile(
                rf"Error tagging repository \({repo_dir}\) with tag \({tag}\):"
                r".*RAN: git"
                r".*STDOUT:"
                r".*mock out"
                r".*STDERR:"
                r".*mock git tag error",
                re.DOTALL
            )
        ):
            ArgoCD._ArgoCD__git_tag_and_push(
                repo_dir=repo_dir,
                tag=tag,
                url=url
            )

            git_mock.push.assert_called_once_with(
                _cwd=repo_dir,
                _out=Any(IOBase)
            )

            git_mock.tag.assert_called_once_with(
                tag,
                '-f',
                _cwd=repo_dir,
                _out=Any(IOBase),
                _err=Any(IOBase)
            )

    @patch.object(sh, 'git')
    def test_ArgoCD__git_tag_and_push_fail_push_tag(self, git_mock):
        repo_dir = '/does/not/matter'
        tag = 'v0.42.0'
        url = None

        git_mock.push.side_effect = [
            create_sh_side_effect(),
            sh.ErrorReturnCode('git', b'mock out', b'mock git push tag error')
        ]

        with self.assertRaisesRegex(
            StepRunnerException,
            re.compile(
                rf"Error pushing tags from repository directory \({repo_dir}\) to"
                rf" repository \({url}\):"
                r".*RAN: git"
                r".*STDOUT:"
                r".*mock out"
                r".*STDERR:"
                r".*mock git push tag error",
                re.DOTALL
            )
        ):
            ArgoCD._ArgoCD__git_tag_and_push(
                repo_dir=repo_dir,
                tag=tag,
                url=url
            )

            git_mock.push.bake().assert_has_calls([
                call(
                    _cwd=repo_dir,
                    _out=Any(IOBase)
                ),
                call(
                    '--tag',
                    _cwd=repo_dir,
                    _out=Any(IOBase)
                )
            ])

            git_mock.tag.assert_called_once_with(
                tag,
                '-f',
                _cwd=repo_dir,
                _out=Any(IOBase),
                _err=Any(IOBase)
            )

    @patch.object(sh, 'git')
    def test_ArgoCD__git_tag_and_push_override_tls(self, git_mock):
        repo_dir = '/does/not/matter'
        tag = 'v0.42.0'
        url = None
        ArgoCD._ArgoCD__git_tag_and_push(
            repo_dir=repo_dir,
            tag=tag,
            url=url,
            force_push_tags=True
        )

        git_mock.push.assert_has_calls([
            call(
                _cwd=repo_dir,
                _out=Any(IOBase)
            ),
            call(
                '--tag',
                '--force',
                _cwd=repo_dir,
                _out=Any(IOBase)
            )
        ])
        git_mock.tag.assert_called_once_with(
            tag,
            '-f',
            _cwd=repo_dir,
            _out=Any(IOBase),
            _err=Any(IOBase)
        )


class TestStepImplementerDeployArgoCD__git_commit_file(TestStepImplementerDeployArgoCDBase):
    @patch.object(sh, 'git')
    def test_ArgoCD__git_commit_file_success(self, git_mock):
        ArgoCD._ArgoCD__git_commit_file(
            git_commit_message='hello world',
            file_path='charts/foo/values-DEV.yaml',
            repo_dir='/does/not/matter'
        )

        git_mock.add.assert_called_once_with(
            'charts/foo/values-DEV.yaml',
            _cwd='/does/not/matter',
            _out=Any(IOBase),
            _err=Any(IOBase)
        )

        git_mock.commit.assert_called_once_with(
            '--allow-empty',
            '--all',
            '--message', 'hello world',
            _cwd='/does/not/matter',
            _out=Any(IOBase),
            _err=Any(IOBase)
        )

    @patch.object(sh, 'git')
    def test_ArgoCD__git_commit_file_fail_add(self, git_mock):
        git_mock.add.side_effect = create_sh_side_effect(
            exception=sh.ErrorReturnCode('git', b'mock out', b'mock git add error')
        )

        with self.assertRaisesRegex(
            StepRunnerException,
            re.compile(
                r"Unexpected error adding file \(charts/foo/values-DEV.yaml\) to commit"
                r" in git repository \(/does/not/matter\):"
                r".*RAN: git"
                r".*STDOUT:"
                r".*mock out"
                r".*STDERR:"
                r".*mock git add error",
                re.DOTALL
            )
        ):
            ArgoCD._ArgoCD__git_commit_file(
                git_commit_message='hello world',
                file_path='charts/foo/values-DEV.yaml',
                repo_dir='/does/not/matter'
            )

            git_mock.add.assert_called_once_with(
                'charts/foo/values-DEV.yaml',
                _cwd='/does/not/matter',
                _out=Any(IOBase),
                _err=Any(IOBase)
            )

            git_mock.commit.assert_not_called()

    @patch.object(sh, 'git')
    def test_ArgoCD__git_commit_file_fail_commit(self, git_mock):
        git_mock.commit.side_effect = create_sh_side_effect(
            exception=sh.ErrorReturnCode('git', b'mock out', b'mock git commit error')
        )

        with self.assertRaisesRegex(
            StepRunnerException,
            re.compile(
                r"Unexpected error commiting file \(charts/foo/values-DEV.yaml\)"
                r" in git repository \(/does/not/matter\):"
                r".*RAN: git"
                r".*STDOUT:"
                r".*mock out"
                r".*STDERR:"
                r".*mock git commit error",
                re.DOTALL
            )
        ):
            ArgoCD._ArgoCD__git_commit_file(
                git_commit_message='hello world',
                file_path='charts/foo/values-DEV.yaml',
                repo_dir='/does/not/matter'
            )

            git_mock.add.assert_called_once_with(
                'charts/foo/values-DEV.yaml',
                _cwd='/does/not/matter',
                _out=Any(IOBase),
                _err=Any(IOBase)
            )

            git_mock.commit.assert_called_once_with(
                '--allow-empty',
                '--all',
                '--message', 'hello world',
                _cwd='/does/not/matter',
                _out=Any(IOBase),
                _err=Any(IOBase)
            )

class TestStepImplementerDeployArgoCD__argocd_sign_in(TestStepImplementerDeployArgoCDBase):
    @patch('sh.argocd', create=True)
    def test_ArgoCD__argocd_sign_in_success_not_insecure(self, argocd_mock):
        argocd_api='argo.dev.ploigos.xyz'
        username='test'
        password='secrettest'
        ArgoCD._ArgoCD__argocd_sign_in(
            argocd_api=argocd_api,
            username=username,
            password=password,
            insecure=False
        )

        argocd_mock.login.assert_called_once_with(
            argocd_api,
            f'--username={username}',
            f'--password={password}',
            None,
            _out=Any(IOBase),
            _err=Any(IOBase)
        )

    @patch('sh.argocd', create=True)
    def test_ArgoCD__argocd_sign_in_success_insecure(self, argocd_mock):
        argocd_api='argo.dev.ploigos.xyz'
        username='test'
        password='secrettest'
        ArgoCD._ArgoCD__argocd_sign_in(
            argocd_api=argocd_api,
            username=username,
            password=password,
            insecure=True
        )

        argocd_mock.login.assert_called_once_with(
            argocd_api,
            f'--username={username}',
            f'--password={password}',
            '--insecure',
            _out=Any(IOBase),
            _err=Any(IOBase)
        )

    @patch('sh.argocd', create=True)
    def test_ArgoCD__argocd_sign_in_fail_not_insecure(self, argocd_mock):
        argocd_api='argo.dev.ploigos.xyz'
        username='test'
        password='secrettest'

        argocd_mock.login.side_effect = create_sh_side_effect(
            exception=sh.ErrorReturnCode('argocd', b'mock out', b'mock login error')
        )

        with self.assertRaisesRegex(
            StepRunnerException,
            re.compile(
                rf"Error logging in to ArgoCD:"
                r".*RAN: argocd"
                r".*STDOUT:"
                r".*mock out"
                r".*STDERR:"
                r".*mock login error",
                re.DOTALL
            )
        ):
            ArgoCD._ArgoCD__argocd_sign_in(
                argocd_api=argocd_api,
                username=username,
                password=password,
                insecure=False
            )

            argocd_mock.login.assert_called_once_with(
                argocd_api,
                f'--username={username}',
                f'--password={password}',
                None,
                _out=Any(IOBase),
                _err=Any(IOBase)
            )

class TestStepImplementerDeployArgoCD__argocd_add_target_cluster(TestStepImplementerDeployArgoCDBase):
    @patch('sh.argocd', create=True)
    def test_ArgoCD__argocd_add_target_cluster_default_cluster(self, argocd_mock):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            step_implementer = self.create_step_implementer(
                step_config={},
                parent_work_dir_path=parent_work_dir_path,
            )

            step_implementer._ArgoCD__argocd_add_target_cluster(
                kube_api='https://kubernetes.default.svc',
                kube_api_skip_tls=False
            )

            argocd_mock.cluster.add.assert_not_called()

    @patch('sh.argocd', create=True)
    def test_ArgoCD__argocd_add_target_cluster_custom_cluster_kube_skip_tls_true(self, argocd_mock):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            step_implementer = self.create_step_implementer(
                step_config={},
                parent_work_dir_path=parent_work_dir_path,
            )
            expected_config_argocd_cluster_context_file_contents = """---
apiVersion: v1
kind: Config
current-context: https://api.dev.ploigos.xyz-context
clusters:
- cluster:
    insecure-skip-tls-verify: true
    server: https://api.dev.ploigos.xyz
  name: default-cluster
contexts:
- context:
    cluster: default-cluster
    user: default-user
  name: https://api.dev.ploigos.xyz-context
preferences:
users:
- name: default-user
  user:
    token: abc123
"""

            step_implementer._ArgoCD__argocd_add_target_cluster(
                kube_api='https://api.dev.ploigos.xyz',
                kube_api_token='abc123',
                kube_api_skip_tls=True
            )

            config_argocd_cluster_context_file_path = os.path.join(
                step_implementer.work_dir_path,
                'config-argocd-cluster-context.yaml'
            )
            argocd_mock.cluster.add.assert_called_once_with(
                '--kubeconfig', config_argocd_cluster_context_file_path,
                'https://api.dev.ploigos.xyz-context',
                _out=Any(IOBase),
                _err=Any(IOBase)
            )

            with open(config_argocd_cluster_context_file_path, 'r') as \
                    config_argocd_cluster_context_file:
                config_argocd_cluster_context_file_contents = \
                    config_argocd_cluster_context_file.read()


                self.assertEqual(
                    config_argocd_cluster_context_file_contents,
                    expected_config_argocd_cluster_context_file_contents
                )

    @patch('sh.argocd', create=True)
    def test_ArgoCD__argocd_add_target_cluster_custom_cluster_kube_skip_tls_false(self, argocd_mock):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            step_implementer = self.create_step_implementer(
                step_config={},
                parent_work_dir_path=parent_work_dir_path,
            )
            expected_config_argocd_cluster_context_file_contents = """---
apiVersion: v1
kind: Config
current-context: https://api.dev.ploigos.xyz-context
clusters:
- cluster:
    insecure-skip-tls-verify: false
    server: https://api.dev.ploigos.xyz
  name: default-cluster
contexts:
- context:
    cluster: default-cluster
    user: default-user
  name: https://api.dev.ploigos.xyz-context
preferences:
users:
- name: default-user
  user:
    token: abc123
"""

            step_implementer._ArgoCD__argocd_add_target_cluster(
                kube_api='https://api.dev.ploigos.xyz',
                kube_api_token='abc123',
                kube_api_skip_tls=False
            )

            config_argocd_cluster_context_file_path = os.path.join(
                step_implementer.work_dir_path,
                'config-argocd-cluster-context.yaml'
            )
            argocd_mock.cluster.add.assert_called_once_with(
                '--kubeconfig', config_argocd_cluster_context_file_path,
                'https://api.dev.ploigos.xyz-context',
                _out=Any(IOBase),
                _err=Any(IOBase)
            )

            with open(config_argocd_cluster_context_file_path, 'r') as \
                    config_argocd_cluster_context_file:
                config_argocd_cluster_context_file_contents = \
                    config_argocd_cluster_context_file.read()


                self.assertEqual(
                    config_argocd_cluster_context_file_contents,
                    expected_config_argocd_cluster_context_file_contents
                )

    @patch('sh.argocd', create=True)
    def test_ArgoCD__argocd_add_target_cluster_fail_custom_cluster_kube_skip_tls_true(self, argocd_mock):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            step_implementer = self.create_step_implementer(
                step_config={},
                parent_work_dir_path=parent_work_dir_path,
            )
            expected_config_argocd_cluster_context_file_contents = """---
apiVersion: v1
kind: Config
current-context: https://api.dev.ploigos.xyz-context
clusters:
- cluster:
    insecure-skip-tls-verify: true
    server: https://api.dev.ploigos.xyz
  name: default-cluster
contexts:
- context:
    cluster: default-cluster
    user: default-user
  name: https://api.dev.ploigos.xyz-context
preferences:
users:
- name: default-user
  user:
    token: abc123
"""

            argocd_mock.cluster.add.side_effect = create_sh_side_effect(
                exception=sh.ErrorReturnCode('argocd', b'mock out', b'mock cluster add error')
            )

            with self.assertRaisesRegex(
                StepRunnerException,
                re.compile(
                    rf"Error adding cluster \(https://api.dev.ploigos.xyz\) to ArgoCD:"
                    r".*RAN: argocd"
                    r".*STDOUT:"
                    r".*mock out"
                    r".*STDERR:"
                    r".*mock cluster add error",
                    re.DOTALL
                )
            ):
                step_implementer._ArgoCD__argocd_add_target_cluster(
                    kube_api='https://api.dev.ploigos.xyz',
                    kube_api_token='abc123',
                    kube_api_skip_tls=True
                )

                config_argocd_cluster_context_file_path = os.path.join(
                    step_implementer.work_dir_path,
                    'config-argocd-cluster-context.yaml'
                )
                argocd_mock.cluster.add.assert_called_once_with(
                    '--kubeconfig', config_argocd_cluster_context_file_path,
                    'https://api.dev.ploigos.xyz-context',
                    _out=Any(IOBase),
                    _err=Any(IOBase)
                )

                with open(config_argocd_cluster_context_file_path, 'r') as \
                        config_argocd_cluster_context_file:
                    config_argocd_cluster_context_file_contents = \
                        config_argocd_cluster_context_file.read()


                    self.assertEqual(
                        config_argocd_cluster_context_file_contents,
                        expected_config_argocd_cluster_context_file_contents
                    )


class TestStepImplementerDeployArgoCD__argocd_app_create_or_update(TestStepImplementerDeployArgoCDBase):
    @patch('sh.argocd', create=True)
    def test__argocd_app_create_or_update_success_sync_auto_no_extra_values_files(self, argocd_mock):
        argocd_app_name = 'test'
        repo = 'https://git.test.xyz'
        revision = 'feature/test'
        path = 'charts/awesome'
        dest_server = 'https://kubernetes.default.svc'
        auto_sync = True
        values_files = []
        ArgoCD._ArgoCD__argocd_app_create_or_update(
            argocd_app_name=argocd_app_name,
            repo=repo,
            revision=revision,
            path=path,
            dest_server=dest_server,
            auto_sync=auto_sync,
            values_files=values_files
        )

        sync_policy = 'automated'
        values_params = None
        argocd_mock.app.create.assert_called_once_with(
            argocd_app_name,
            f'--repo={repo}',
            f'--revision={revision}',
            f'--path={path}',
            f'--dest-server={dest_server}',
            f'--dest-namespace={argocd_app_name}',
            f'--sync-policy={sync_policy}',
            values_params,
            '--upsert',
            _out=Any(IOBase),
            _err=Any(IOBase)
        )

    @patch('sh.argocd', create=True)
    def test__argocd_app_create_or_update_success_sync_none_no_extra_values_files(self, argocd_mock):
        argocd_app_name = 'test'
        repo = 'https://git.test.xyz'
        revision = 'feature/test'
        path = 'charts/awesome'
        dest_server = 'https://kubernetes.default.svc'
        auto_sync = False
        values_files = []
        ArgoCD._ArgoCD__argocd_app_create_or_update(
            argocd_app_name=argocd_app_name,
            repo=repo,
            revision=revision,
            path=path,
            dest_server=dest_server,
            auto_sync=auto_sync,
            values_files=values_files
        )

        sync_policy = 'none'
        values_params = None
        argocd_mock.app.create.assert_called_once_with(
            argocd_app_name,
            f'--repo={repo}',
            f'--revision={revision}',
            f'--path={path}',
            f'--dest-server={dest_server}',
            f'--dest-namespace={argocd_app_name}',
            f'--sync-policy={sync_policy}',
            values_params,
            '--upsert',
            _out=Any(IOBase),
            _err=Any(IOBase)
        )

    @patch('sh.argocd', create=True)
    def test__argocd_app_create_or_update_success_sync_auto_1_values_files(self, argocd_mock):
        argocd_app_name = 'test'
        repo = 'https://git.test.xyz'
        revision = 'feature/test'
        path = 'charts/awesome'
        dest_server = 'https://kubernetes.default.svc'
        auto_sync = True
        values_files = ['values-foo.yaml']
        ArgoCD._ArgoCD__argocd_app_create_or_update(
            argocd_app_name=argocd_app_name,
            repo=repo,
            revision=revision,
            path=path,
            dest_server=dest_server,
            auto_sync=auto_sync,
            values_files=values_files
        )

        sync_policy = 'automated'
        values_params = ['--values=values-foo.yaml']
        argocd_mock.app.create.assert_called_once_with(
            argocd_app_name,
            f'--repo={repo}',
            f'--revision={revision}',
            f'--path={path}',
            f'--dest-server={dest_server}',
            f'--dest-namespace={argocd_app_name}',
            f'--sync-policy={sync_policy}',
            values_params,
            '--upsert',
            _out=Any(IOBase),
            _err=Any(IOBase)
        )

    @patch('sh.argocd', create=True)
    def test__argocd_app_create_or_update_success_sync_auto_2_values_files(self, argocd_mock):
        argocd_app_name = 'test'
        repo = 'https://git.test.xyz'
        revision = 'feature/test'
        path = 'charts/awesome'
        dest_server = 'https://kubernetes.default.svc'
        auto_sync = True
        values_files = ['values-foo.yaml', 'values-DEV.yaml']
        ArgoCD._ArgoCD__argocd_app_create_or_update(
            argocd_app_name=argocd_app_name,
            repo=repo,
            revision=revision,
            path=path,
            dest_server=dest_server,
            auto_sync=auto_sync,
            values_files=values_files
        )

        sync_policy = 'automated'
        values_params = ['--values=values-foo.yaml', '--values=values-DEV.yaml']
        argocd_mock.app.create.assert_called_once_with(
            argocd_app_name,
            f'--repo={repo}',
            f'--revision={revision}',
            f'--path={path}',
            f'--dest-server={dest_server}',
            f'--dest-namespace={argocd_app_name}',
            f'--sync-policy={sync_policy}',
            values_params,
            '--upsert',
            _out=Any(IOBase),
            _err=Any(IOBase)
        )

    @patch('sh.argocd', create=True)
    def test__argocd_app_create_or_update_fail_sync_auto_1_values_files(self, argocd_mock):
        argocd_mock.app.create.side_effect = create_sh_side_effect(
            exception=sh.ErrorReturnCode('argocd', b'mock out', b'mock create error')
        )

        argocd_app_name = 'test'
        repo = 'https://git.test.xyz'
        revision = 'feature/test'
        path = 'charts/awesome'
        dest_server = 'https://kubernetes.default.svc'
        auto_sync = True
        values_files = ['values-foo.yaml']

        with self.assertRaisesRegex(
            StepRunnerException,
            re.compile(
                rf"Error creating or updating ArgoCD app \({argocd_app_name}\):"
                r".*RAN: argocd"
                r".*STDOUT:"
                r".*mock out"
                r".*STDERR:"
                r".*mock create error",
                re.DOTALL
            )
        ):
            ArgoCD._ArgoCD__argocd_app_create_or_update(
                argocd_app_name=argocd_app_name,
                repo=repo,
                revision=revision,
                path=path,
                dest_server=dest_server,
                auto_sync=auto_sync,
                values_files=values_files
            )

        sync_policy = 'automated'
        values_params = ['--values=values-foo.yaml']
        argocd_mock.app.create.assert_called_once_with(
            argocd_app_name,
            f'--repo={repo}',
            f'--revision={revision}',
            f'--path={path}',
            f'--dest-server={dest_server}',
            f'--dest-namespace={argocd_app_name}',
            f'--sync-policy={sync_policy}',
            values_params,
            '--upsert',
            _out=Any(IOBase),
            _err=Any(IOBase)
        )

class TestStepImplementerDeployArgoCD__argocd_app_sync(TestStepImplementerDeployArgoCDBase):
    @patch('sh.argocd', create=True)
    def test__argocd_app_sync_success(self, argocd_mock):
        ArgoCD._ArgoCD__argocd_app_sync(
            argocd_app_name='test',
            argocd_sync_timeout_seconds=120,
            argocd_sync_retry_limit=3
        )

        argocd_mock.app.sync.assert_called_once_with(
            '--prune',
            '--timeout', 120,
            '--retry-limit', 3,
            'test',
            _out=Any(IOBase),
            _err=Any(IOBase)
        )
        argocd_mock.app.wait.assert_called_once_with(
            '--timeout', 120,
            '--health',
            'test',
            _out=Any(IOBase),
            _err=Any(IOBase)
        )

    @patch('sh.argocd', create=True)
    def test__argocd_app_sync_success_retry(self, argocd_mock):
        ArgoCD._ArgoCD__argocd_app_sync(
            argocd_app_name='test',
            argocd_sync_timeout_seconds=120,
            argocd_sync_retry_limit=4
        )

        argocd_mock.app.sync.assert_called_once_with(
            '--prune',
            '--timeout', 120,
            '--retry-limit', 4,
            'test',
            _out=Any(IOBase),
            _err=Any(IOBase)
        )
        argocd_mock.app.wait.assert_called_once_with(
            '--timeout', 120,
            '--health',
            'test',
            _out=Any(IOBase),
            _err=Any(IOBase)
        )

    @patch('sh.argocd', create=True)
    def test__argocd_app_sync_fail_sync(self, argocd_mock):
        argocd_mock.app.sync.side_effect = create_sh_side_effect(
            exception=sh.ErrorReturnCode('argocd', b'mock out', b'mock sync error')
        )

        with self.assertRaisesRegex(
            StepRunnerException,
            re.compile(
                r"Error synchronization ArgoCD Application \(test\):"
                r".*RAN: argocd"
                r".*STDOUT:"
                r".*mock out"
                r".*STDERR:"
                r".*mock sync error",
                re.DOTALL
            )
        ):
            ArgoCD._ArgoCD__argocd_app_sync(
                argocd_app_name='test',
                argocd_sync_timeout_seconds=120,
                argocd_sync_retry_limit=3
            )

        argocd_mock.app.sync.assert_called_once_with(
            '--prune',
            '--timeout', 120,
            '--retry-limit', 3,
            'test',
            _out=Any(IOBase),
            _err=Any(IOBase)
        )
        argocd_mock.app.wait.assert_not_called()

    @patch('sh.argocd', create=True)
    def test__argocd_app_sync_fail_wait(self, argocd_mock):
        argocd_mock.app.wait.side_effect = create_sh_side_effect(
            exception=sh.ErrorReturnCode('argocd', b'mock out', b'mock wait error')
        )

        with self.assertRaisesRegex(
            StepRunnerException,
            re.compile(
                r"Error waiting for ArgoCD Application \(test\) synchronization:"
                r".*RAN: argocd"
                r".*STDOUT:"
                r".*mock out"
                r".*STDERR:"
                r".*mock wait error",
                re.DOTALL
            )
        ):
            ArgoCD._ArgoCD__argocd_app_sync(
                argocd_app_name='test',
                argocd_sync_timeout_seconds=120,
                argocd_sync_retry_limit=3
            )

        argocd_mock.app.sync.assert_called_once_with(
            '--prune',
            '--timeout', 120,
            '--retry-limit', 3,
            'test',
            _out=Any(IOBase),
            _err=Any(IOBase)
        )
        argocd_mock.app.wait.assert_called_once_with(
            '--timeout', 120,
            '--health',
            'test',
            _out=Any(IOBase),
            _err=Any(IOBase)
        )

class TestStepImplementerDeployArgoCD__argocd_get_app_manifest(TestStepImplementerDeployArgoCDBase):
    @patch('sh.argocd', create=True)
    def test___argocd_get_app_manifest_success_live(self, argocd_mock):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            step_implementer = self.create_step_implementer(
                step_config={},
                parent_work_dir_path=parent_work_dir_path,
            )

            argocd_app_manifest_file = step_implementer._ArgoCD__argocd_get_app_manifest(
                argocd_app_name='test',
                source='live'
            )

            self.assertIsNotNone(argocd_app_manifest_file)
            argocd_mock.app.manifests.assert_called_once_with(
                '--source=live',
                'test',
                _out=Any(IOBase),
                _err=Any(IOBase)
            )

    @patch('sh.argocd', create=True)
    def test___argocd_get_app_manifest_success_git(self, argocd_mock):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            step_implementer = self.create_step_implementer(
                step_config={},
                parent_work_dir_path=parent_work_dir_path,
            )

            argocd_app_manifest_file = step_implementer._ArgoCD__argocd_get_app_manifest(
                argocd_app_name='test',
                source='git'
            )

            self.assertIsNotNone(argocd_app_manifest_file)
            argocd_mock.app.manifests.assert_called_once_with(
                '--source=git',
                'test',
                _out=Any(IOBase),
                _err=Any(IOBase)
            )

    @patch('sh.argocd', create=True)
    def test___argocd_get_app_manifest_fail(self, argocd_mock):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')

            step_implementer = self.create_step_implementer(
                step_config={},
                parent_work_dir_path=parent_work_dir_path,
            )

            argocd_mock.app.manifests.side_effect = create_sh_side_effect(
                exception=sh.ErrorReturnCode('argocd', b'mock out', b'mock error')
            )

            with self.assertRaisesRegex(
                StepRunnerException,
                re.compile(
                    r"Error reading ArgoCD Application \(invalid\) manifest:"
                    r".*RAN: argocd"
                    r".*STDOUT:"
                    r".*mock out"
                    r".*STDERR:"
                    r".*mock error",
                    re.DOTALL
                )
            ):
                argocd_app_manifest_file = step_implementer._ArgoCD__argocd_get_app_manifest(
                    argocd_app_name='invalid',
                    source='live'
                )

                self.assertIsNotNone(argocd_app_manifest_file)
                argocd_mock.app.manifests.assert_called_once_with(
                    '--source=live',
                    'invalid',
                    _out=Any(IOBase),
                    _err=Any(IOBase)
                )



import os

import unittest
from testfixtures import TempDirectory

from tests.helpers.base_test_case import BaseTestCase

from ploigos_step_runner.utils.dict import deep_merge

class TestDictUtils(BaseTestCase):
    def test_deep_merge_no_conflict(self):
        dict1 = {
            'step-runner-config': {
                'step-foo': {
                    'implementer': 'foo1',
                    'config': {
                        'test0': 'foo',
                        'test1': 'foo'
                    }
                }
            }
        }

        dict2 = {
            'step-runner-config': {
                'step-foo': {
                    'implementer': 'foo1',
                    'config': {
                        'test2': 'bar'
                    }
                }
            }
        }

        result = deep_merge(dict1, dict2)

        # assert that it is in an inplace deep merge of the first param
        self.assertEqual(result, dict1)

        # assert expected merge result
        self.assertEqual(result, {
            'step-runner-config': {
                'step-foo': {
                    'implementer': 'foo1',
                    'config': {
                        'test0': 'foo',
                        'test1': 'foo',
                        'test2': 'bar'
                    }
                }
            }
        })

    def test_deep_merge_conflict_no_overwrite(self):
        dict1 = {
            'step-runner-config': {
                'step-foo': {
                    'implementer': 'foo1',
                    'config': {
                        'test0': 'foo',
                        'test1': 'foo'
                    }
                }
            }
        }

        dict2 = {
            'step-runner-config': {
                'step-foo': {
                    'implementer': 'foo1',
                    'config': {
                        'test1': 'bar',
                        'test2': 'bar'
                    }
                }
            }
        }

        with self.assertRaisesRegex(
                ValueError,
                r"Conflict at step-runner-config.step-foo.config.test1"):

            deep_merge(dict1, dict2)

    def test_deep_merge_conflict_overwrite_duplicate_keys(self):
        dict1 = {
            'step-runner-config': {
                'step-foo': {
                    'implementer': 'foo1',
                    'config': {
                        'test0': 'foo',
                        'test1': 'foo'
                    }
                }
            }
        }

        dict2 = {
            'step-runner-config': {
                'step-foo': {
                    'implementer': 'foo1',
                    'config': {
                        'test1': 'bar',
                        'test2': 'bar'
                    }
                }
            }
        }

        result = deep_merge(
            dest=dict1,
            source=dict2,
            overwrite_duplicate_keys=True
        )

        # assert expected merge result
        self.assertEqual(result, {
            'step-runner-config': {
                'step-foo': {
                    'implementer': 'foo1',
                    'config': {
                        'test0': 'foo',
                        'test1': 'bar',
                        'test2': 'bar'
                    }
                }
            }
        })



"""`StepImplementer` for the `generate-metadata` step using Git.

Step Configuration
------------------
Step configuration expected as input to this step.
Could come from:

  * static configuration
  * runtime configuration
  * previous step results

Configuration Key     | Required? | Default | Description
----------------------|-----------|---------|-----------
`repo-root`           | Yes       | `./`    | Directory path to the Git repo to generate \
                                            metadata from. Must not be in a detached head state \
                                            so that `pre-release` value can be determined.
`build-string-length` | Yes       | `7`     | Length of the Git hash to use for the \
                                              `build` portion of the semantic version.

Result Artifacts
----------------
Results artifacts output by this step.

Result Artifact Key | Description
--------------------|------------
`pre-release` | Value to use for `pre-release` portion of semantic version \
                (https://semver.org/). Uses the Git branch name.
`build`       | Value to use for `build` portion of semantic version (https://semver.org/). \
                Uses a portion of the latest Git commit hash.
"""

import re

from git import InvalidGitRepositoryError, Repo

from ploigos_step_runner import StepImplementer, StepResult

DEFAULT_CONFIG = {
    'repo-root': './',
    'build-string-length': 7
}

REQUIRED_CONFIG_OR_PREVIOUS_STEP_RESULT_ARTIFACT_KEYS = [
    'repo-root',
    'build-string-length'
]

class Git(StepImplementer):  # pylint: disable=too-few-public-methods
    """
    StepImplementer for the generate-metadata step for Git.
    """

    @staticmethod
    def step_implementer_config_defaults():
        """Getter for the StepImplementer's configuration defaults.

        Returns
        -------
        dict
            Default values to use for step configuration values.

        Notes
        -----
        These are the lowest precedence configuration values.

        """
        return DEFAULT_CONFIG

    @staticmethod
    def _required_config_or_result_keys():
        """Getter for step configuration or previous step result artifacts that are required before
        running this step.

        See Also
        --------
        _validate_required_config_or_previous_step_result_artifact_keys

        Returns
        -------
        array_list
            Array of configuration keys or previous step result artifacts
            that are required before running the step.
        """
        return REQUIRED_CONFIG_OR_PREVIOUS_STEP_RESULT_ARTIFACT_KEYS

    def _run_step(self):
        """Runs the step implemented by this StepImplementer.

        Returns
        -------
        StepResult
            Object containing the dictionary results of this step.
        """
        step_result = StepResult.from_step_implementer(self)

        repo_root = self.get_value('repo-root')
        build_string_length = self.get_value('build-string-length')

        try:
            repo = Repo(repo_root)
        except InvalidGitRepositoryError:
            step_result.success = False
            step_result.message = 'Given directory (repo_root) is not a Git repository'
            return step_result

        if repo.bare:
            step_result.success = False
            step_result.message = 'Given directory (repo_root) is a bare Git repository'
            return step_result

        # The SemanticVersion StepImplementer uses the branch name (as stored by 'pre-release'
        # below), so this step requires the git repository to not be in a detached head state. If
        # there are any brilliant ideas for specifying an appropriate pre-release value while in a
        # detached head state, pull requests are welcome!
        if repo.head.is_detached:
            step_result.success = False
            step_result.message = 'Expected a Git branch in given directory (repo_root) but has' \
                                  ' a detached head'
            return step_result

        git_branch = str(repo.head.reference)
        pre_release_regex = re.compile(r"/", re.IGNORECASE)
        pre_release = re.sub(pre_release_regex, '_', git_branch)
        step_result.add_artifact(
            name='pre-release',
            value=pre_release
        )

        try:
            git_branch_last_commit_hash = str(repo.head.reference.commit)[:build_string_length]

            step_result.add_artifact(
                name='build',
                value=git_branch_last_commit_hash
            )
        except ValueError:
            step_result.success = False
            step_result.message = 'Given directory (repo_root) is a git branch (git_branch) with' \
                                  ' no commit history'
            return step_result

        return step_result



"""Shared utilities for dealing with IO
"""

import io
import random
import re


def create_sh_redirect_to_multiple_streams_fn_callback(streams):
    """Creates and returns a function callback that will write given data to multiple given streams.

    AKA: this essentially allows you to do 'tee' for sh commands.

    Parameters
    ----------
    streams : list of io.IOBase
        Streams to write to.

    Examples
    --------
    Will write output directed at stdout to stdout and a results file and output directed
    at stderr to stderr and a results file.
    >>> with open('/tmp/results_file', 'w') as results_file:
    ...     out_callback = create_sh_redirect_to_multiple_streams_fn_callback([
    ...         sys.stdout,
    ...         results_file
    ...     ])
    ...     err_callback = create_sh_redirect_to_multiple_streams_fn_callback([
    ...         sys.stderr,
    ...         results_file
    ...     ])
    ...     sh.echo('hello world')
    hello world

    Returns
    -------
    function(data)
        Function that takes one parameter, data, and writes that value to all the given streams.
    """

    def sh_redirect_to_multiple_streams(data):
        for stream in streams:
            stream.write(data)

    return sh_redirect_to_multiple_streams


class TextIOSelectiveObfuscator(io.TextIOBase):
    """Extends the base class for text streams to allow the obfuscation of given patterns.

    This is useful to prevent accidentally writing "sensitive" information to stdout/stderr.

    Parameters
    ----------
    parent_stream : IOBase
        IO stream to write to after obfuscating any text written to this stream.
    randomize_replacment_length : bool, optional
        True to randomize the length of the text being obfuscated in the stream.
        False to use the same length replacement for any obfuscated text in the stream.
    replacement_char : char
        Character to replace the target strings to obfuscate with.

    Attributes
    ----------
    __parent_stream : IOBase
    __obfuscation_patterns : list
    __replacement_char : char
    __randomize_replacement_length : bool
    __random_replacement_length_min : int
    __random_replacement_length_max : int
    """

    def __init__(self, parent_stream, randomize_replacment_length=True, replacement_char='*'):
        self.__parent_stream = parent_stream
        self.__obfuscation_patterns = []
        self.__replacement_char = replacement_char
        self.__randomize_replacement_length = randomize_replacment_length
        self.__random_replacement_length_min = 5
        self.__random_replacement_length_max = 40
        super().__init__()

    @property
    def parent_stream(self):
        """
        Returns
        -------
        IOBase
            IO stream this stream writes to after obfuscating any text written to this stream.
        """
        return self.__parent_stream

    @property
    def replacement_char(self):
        """
        Returns
        -------
        char
            Character to replace the target strings to obfuscate with.
        """
        return self.__replacement_char

    @replacement_char.setter
    def replacement_char(self, replacement_char):
        """
        Parameters
        ----------
        replacement_char : char
            Character to replace the target strings to obfuscate with.
        """
        self.__replacement_char = replacement_char

    @property
    def randomize_replacement_length(self):
        """
        Returns
        -------
        bool
            True if this stream is randomizing the length of the text being obfuscated.
            False if this stream is using the same length replacement for any obfuscated text.
        """
        return self.__randomize_replacement_length

    def add_obfuscation_targets(self, targets):
        """Adds a target pattern to be obfuscated whenever writing to this stream.

        Notes
        -----
        This is a bit involved to deal with secrets that span multiple lines and various ways they
        can be printed. so the regex gets pretty involved to escape the right things and ignore
        whitespace, so forth and so on.

        There are unit tests covering the scenarios this is dealing with, if you are messing in
        here be sure you don't break any of the existing unit tests.

        Parameters
        ----------
        targets : list or pattern
            The target patterns to be obfuscated when writing to this stream.
        """
        if not isinstance(targets, list):
            targets = [targets]

        for target in targets:
            target_pattern = target

            # replace any amount of whitespace with a single space
            target_pattern = re.sub(r'\s+', ' ', target_pattern)

            # strip off leading and trialing whitespace
            target_pattern = target_pattern.strip()

            # escape for use in regex pattern
            target_pattern = re.escape(target_pattern)

            # the spaces we added in now got escaped, so unescape them and turn them into .*
            target_pattern = re.sub(r'\\ ', r'.*', target_pattern)

            # eat up pre and post newlines
            # target_pattern = f"(\s*)({target_pattern})(\s*)"

            # compile the pattern for re-use and make sure that .* matches accross lines
            target_compiled_pattern = re.compile(target_pattern, re.DOTALL)

            # add the pattern
            self.__obfuscation_patterns.append(target_compiled_pattern)

    def __obfuscator(self, match):
        """Given a regex match returns a corresponding obfuscated string.

        Parameters
        ----------
        match : re.Match
            re.Match to replace with obfuscated text

        Returns
        -------
        str
            String to replace the given match with.

        Also See
        --------
        re.sub
        """

        if self.randomize_replacement_length:
            replacement_length = random.randint(
                self.__random_replacement_length_min,
                self.__random_replacement_length_max
            )
        else:
            replacement_length = len(match.group())

        return self.replacement_char * replacement_length

    def write(self, given):
        """Writes to this streams parent stream after obfuscating all of the obfuscation targets.

        Parameters
        ----------
        given : str
            Given string to write to the parent stream after obfuscating.

        Returns
        -------
        int
            Number of characters written.

        See Also
        --------
        io.TextIOBase.write
        """

        if isinstance(given, bytes):
            obfuscated = given.decode('utf-8')
        else:
            obfuscated = given

        for obfuscation_pattern in self.__obfuscation_patterns:
            obfuscated = obfuscation_pattern.sub(self.__obfuscator, obfuscated)

        return self.parent_stream.write(obfuscated)

    def flush(self):
        """Flush the parent stream.

        See Also
        --------
        io.TextIOBase.flush
        """
        self.parent_stream.flush()


class TextIOIndenter(io.TextIOBase):
    """Adds an indent to the first string written and after every new line written to this stream.

    Notes
    -----
    WARNING:    There will be a dangling indent after the last new line written to this stream
                because there is no "good" way of knowing the last new line ever written to this
                stream.

    Parameters
    ----------
    parent_stream : IOBase
        Stream to write to after indenting what is written to this stream
    indent_level : int, optional
        Level to indent to.
        Will be multiplied by the indent_size.
    indent_size : int, optional
        Size of each indent.
        Will be multipled by the indent_level.
    indent_char : str, optional
        Character to use for indent.
        Will be multiplied by the indent_size and the ident_level and prepended before
        each line written to this stream.
    """

    def __init__(self, parent_stream, indent_level=0, indent_size=4, indent_char=' '):
        self.__parent_stream = parent_stream
        self.__indent_level = indent_level
        self.__indent_size = indent_size
        self.__indent_char = indent_char
        self.__unwritten_to = True
        super().__init__()

    @property
    def parent_stream(self):
        """Returns parent stream that this stream wraps

        Returns
        -------
        IOBase
            Stream to write to after indenting what is written to this stream
        """
        return self.__parent_stream

    @property
    def indent_level(self):
        """Level to indent to.

        Returns
        -------
        int
            Level to indent to.
        """
        return self.__indent_level

    @property
    def indent_size(self):
        """Size of each indent.

        Returns
        -------
        int
            Size of each indent.
        """
        return self.__indent_size

    @property
    def indent_char(self):
        """Character to use for indent.

        Will be multiplied by the indent_size and the ident_level and prepended before
        each line written to this stream.

        Returns
        -------
        str
            Character to use for indent.
        """
        return self.__indent_char

    def write(self, given):
        """Indents the begining of the given text as well as every new line in the given text
        and then writes it to this steams parent_stream.

        Notes
        -----
        If there is a new line at the end of the given string there will not be an indent
        written after that new line assuming that the next line will be written by this stream
        and therefor get indented as being initially written to this stream.

        Parameters
        ----------
        given : str or bytes (utf-8)
            String to indent every line of before writing to parent stream.

        Examples
        --------
        >>> TextIOIndenter(sys.stdout).write("hello world\\n")
        hello world
        <BLANKLINE>

        >>> TextIOIndenter(sys.stdout, 1).write("hello world\\n")
            hello world
            <BLANKLINE>

        >>> TextIOIndenter(sys.stdout, 2).write("hello world\\n")
                hello world
                <BLANKLINE>

        >>> TextIOIndenter(sys.stdout).write("\\nhello world\\n")
        <BLANKLINE>
        hello world
        <BLANKLINE>

        >>> TextIOIndenter(sys.stdout, 1).write("\\nhello world\\n")
            <BLANKLINE>
            hello world
            <BLANKLINE>

        >>> TextIOIndenter(sys.stdout, 2).write("\\nhello world\\n")
                <BLANKLINE>
                hello world
                <BLANKLINE>

        >>> TextIOIndenter(sys.stdout, 0, 2, '-').write("hello world")
        hello world

        >>> TextIOIndenter(sys.stdout, 1, 2, '-').write("hello world")
        --hello world

        >>> TextIOIndenter(sys.stdout, 2, 2, '-').write(hello world")
        ----hello world

        >>> TextIOIndenter(sys.stdout, 1).write("hello\\nworld\\n")
            hello
            world
            <BLANKLINE>


        >>> TextIOIndenter(sys.stdout, 2).write("hello\\nworld\\n")
                hello
                world
                <BLANKLINE>

        >>> indenter = TextIOIndenter(sys.stdout, 1)
        ... indenter.write("hello\\nworld\\n")
        ... indenter.write("foo bar\\n")
            hello
            world
            foo bar

        >>> indenter = TextIOIndenter(sys.stdout, 1)
        ... indenter.write("hello world ")
        ... indenter.write("foo bar\\n")
        ... indenter.write("this is a test, ")
        ... indenter.write("more testing\\n")
        ... indenter.write("fortytwo\\n")
            hello world foo bar
            this is a test more testing
            fortytwo

        Returns
        -------
        int
            Number of characters written.

        See Also
        --------
        io.TextIOBase.write
        """
        if isinstance(given, bytes):
            indented = given.decode('utf-8')
        else:
            indented = given

        # create the indent to insert at begining of given text and every new line in that text
        indent_chars = self.indent_char * (self.indent_size * self.indent_level)

        if self.__unwritten_to:
            self.__unwritten_to = False
            indented = f"{indent_chars}{indented}"

        # add indent after every new line
        # NOTE: \1 is capture group one and contains the original new line character
        indented = re.sub(r"(\r\n|\r|\n)", r"\1" + indent_chars, indented)

        return self.parent_stream.write(indented)

    def flush(self):
        """Flush the parent stream.

        See Also
        --------
        io.TextIOBase.flush
        """
        self.parent_stream.flush()



import os
from io import IOBase
from pathlib import Path
from unittest.mock import patch

import sh
from shutil import copy
from ploigos_step_runner import StepResult, WorkflowResult
from ploigos_step_runner.config.config import Config
from ploigos_step_runner.step_implementers.shared import RekorSignGeneric
from testfixtures import TempDirectory
from tests.helpers.base_step_implementer_test_case import \
    BaseStepImplementerTestCase
from tests.helpers.test_utils import Any
from ploigos_step_runner.utils.pgp import detach_sign_with_pgp_key
from ploigos_step_runner.utils.pgp import import_pgp_key
from ploigos_step_runner.utils.pgp import export_pgp_public_key


class TestStepImplementerSharedRekorSignGeneric(BaseStepImplementerTestCase):
    TEST_REKOR_ENTRY = {
        "kind": "rekord",
        "apiVersion": "0.0.1",
        "spec": {
            "signature": {
                "format": "pgp",
                "content": "",
                "publicKey": {
                    "content": ""
                }
            },
            "data": {
                "content": "ewogICAgInN0ZXAtcnVubmVyLXJlc3VsdHMiOiB7fQp9",
                "hash": {
                    "algorithm": "sha256",
                    "value": "7e9bb92be27b8897c4be4e2b5a185a0823bd4b1f566ef81312a96bbc278cf716"
                }
            },
            "extraData": {
                    "signed-artifact-uri": "http://foo.bar/artifact_to_sign"
            }
        }
    }

    TEST_REKOR_UUID = 'b08416d417acdb0610d4a030d8f697f9d0a718024681a00fa0b9ba67072a38b5'
    TEST_REKOR_SERVER_URL = 'http://rekor.apps.tssc.rht-set.com'
    TEST_ARTIFACT_TO_SIGN_URI_CONFIG_KEY = 'test-key'
    TEST_ARTIFACT_TO_SIGN_URI = 'http://foo.bar/artifact_to_sign'
    TEST_PGP_SIGNER_PRIVATE_KEY = ""
    TEST_REKOR_ENTRY_URI='http://rekor.apps.tssc.rht-set.com/api/v1/log/entries/' + TEST_REKOR_UUID

    TEST_FAKE_FINGERPRINT = 'DEB72DAF711EF3586266C98D9D65E8ECADD724B4'

    TEST_FAKE_PUBLIC_KEY = '''\
-----BEGIN PGP PUBLIC KEY BLOCK-----

mQENBGDKRgEBCACyVB9ZZodlYBf5Fw5iWCohDyjFkypQjRFVaQz8XbvOzrU0reNy
fQpXFzEiNcGxhUHDM538Zv/adIOWMj3EvFTEuCY9I5DSfi4A8AfjSxLMUVSJ85f9
rmHWKoCKoadF7FduKWnGEUk4woVNmhoVqpNMbV6IczNX2vmr7omx4/KgLu4k8WeW
3C0ItJES1fTz9epfynpndXZ9+Iiqiyx8d2cWcVtodv60bJkVIGdLhvyn3/vfSFvf
iSkw4FcKjpXt0k7Kd0e2mv5PHqEYWtkN41p6hDClu3i4EgnFwRjJK9H7lUPcttIJ
hS44e5HhLrihPVK91HZ5Eh0k5jpgQpkNV+ntABEBAAG0O0pvaG4gVGhvbXBraW5z
IChSZWQgSGF0KSAoTm8gY29tbWVudCkgPGp0aG9tcGtpQHJlZGhhdC5jb20+iQFO
BBMBCAA4FiEE3rctr3Ee81hiZsmNnWXo7K3XJLQFAmDKRgECGwMFCwkIBwIGFQoJ
CAsCBBYCAwECHgECF4AACgkQnWXo7K3XJLQANAgAjp5GQALhmhaxLXQJs9tmLzgB
VLOTn6r3z6emlIjjxZYlwLCV+xXZ12Igg+1L5IJo0Nrk2xmcAgp+3WScx29xtkbW
j44jt7eghH4clkSau8RTJWQlyMyzmLVGEDgqBr+YICcnV85NS31ILigvMorKgszP
RVJpG9G+GQRNm7L17sQUSYItX/4ZAbzBr/RhD4IRkgwFxlgrsVujcGN4UMQ2f+uy
b7jY7Gx4o3lKZc+2rFAp6Z9xkChdl69UzJA9Cb7MGNPDE2rFItqXQRUX+4bVOWjy
ffPOuPFqVZJ4989lAfNFh5zitVs2aYGMuWbrRp/AMqBY+2AWw+78/Gv2b6iPfbkB
DQRgykYBAQgAxYk3SS3wk3vGjGF8qAQ6AnMnU2XpqlGt9uTJnChDyzhb964Yxx6+
vN3EqChoAROjyaUQN2GzmUezm1kNDa9sw4KE+a50d5nJqEniz4UvUUTzi9FluTb5
EkRGr1zP0rd1PPWvMB9gil2lie4d7O9iwYT5KUDhiEB+LH8GvI0YENcYDN46CBvG
Qb89qhttXYp22DIll3mQPoabdmWFYD1xYRGoEwP/Zjkx+9MPOASStD4oaULkHoDn
qNMQNrVEY9YQ6JrG55a2XTLEcBaWmLyGnytRjVSs0xp8fzZCwTDgCkcoyAnNrQ7k
wiJJV3fvHP5i6meEdPdRSR4Bwc173ZzUdQARAQABiQE2BBgBCAAgFiEE3rctr3Ee
81hiZsmNnWXo7K3XJLQFAmDKRgECGwwACgkQnWXo7K3XJLS24AgAoGke/g08/FEz
Fml1e81zYEB6+GyggQS365JA16Ev+nvOQbjEk55PvZFfNMbsI0T7IamweX3BzlPf
Sb8HZsjJp0SNlhu7I1P4X3nMa2OofSEE8qd3ptEmAWqewhW+DqGDA23ccAO+VDSj
UvF+nhrYFM/BJY/SE5QyDcDRuJUG3hEAN0FPGSEbWZoAgPpQKdmwLwHeTtdadvAH
NNqw3gdbZjlF+RQbggHc4sXtK/H7wk3SU501BgP10GkfsFUekGiAlkAmYAO6gUnQ
1cfRMdTTzuGVc6j/8VshK1aYRR5A3g5rP4ZbNReqQddZmB20qe4ekkFKbRvHX/Em
B8pBNt1QOA==
=u6VF
-----END PGP PUBLIC KEY BLOCK-----
'''

    TEST_FAKE_PRIVATE_KEY = '''
        -----BEGIN RSA PRIVATE KEY-----
        MIICXAIBAAKBgQCqGKukO1De7zhZj6+H0qtjTkVxwTCpvKe4eCZ0FPqri0cb2JZfXJ/DgYSF6vUp
        wmJG8wVQZKjeGcjDOL5UlsuusFncCzWBQ7RKNUSesmQRMSGkVb1/3j+skZ6UtW+5u09lHNsj6tQ5
        1s1SPrCBkedbNf0Tp0GbMJDyR4e9T04ZZwIDAQABAoGAFijko56+qGyN8M0RVyaRAXz++xTqHBLh
        3tx4VgMtrQ+WEgCjhoTwo23KMBAuJGSYnRmoBZM3lMfTKevIkAidPExvYCdm5dYq3XToLkkLv5L2
        pIIVOFMDG+KESnAFV7l2c+cnzRMW0+b6f8mR1CJzZuxVLL6Q02fvLi55/mbSYxECQQDeAw6fiIQX
        GukBI4eMZZt4nscy2o12KyYner3VpoeE+Np2q+Z3pvAMd/aNzQ/W9WaI+NRfcxUJrmfPwIGm63il
        AkEAxCL5HQb2bQr4ByorcMWm/hEP2MZzROV73yF41hPsRC9m66KrheO9HPTJuo3/9s5p+sqGxOlF
        L0NDt4SkosjgGwJAFklyR1uZ/wPJjj611cdBcztlPdqoxssQGnh85BzCj/u3WqBpE2vjvyyvyI5k
        X6zk7S0ljKtt2jny2+00VsBerQJBAJGC1Mg5Oydo5NwD6BiROrPxGo2bpTbu/fhrT8ebHkTz2epl
        U9VQQSQzY1oZMVX8i1m5WUTLPz2yLJIBQVdXqhMCQBGoiuSoSjafUhV7i1cEGpb88h5NBYZzWXGZ
        37sJ5QsW+sJyoNde3xH8vdXhzU7eT82D6X/scw9RZz+/6rCJ4p0=
        -----END RSA PRIVATE KEY-----'''

    def create_step_implementer(
            self,
            step_config={},
            workflow_result=None,
            parent_work_dir_path='',
            artifact_to_sign_uri_config_key=None
    ):

        step_implementer = self.create_given_step_implementer_rekor_sign_generic(
            step_implementer=RekorSignGeneric,
            step_config=step_config,
            step_name='automated-governance',
            implementer='RekorSignGeneric',
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path,
            artifact_to_sign_uri_config_key=artifact_to_sign_uri_config_key
            )
        return step_implementer

    def create_given_step_implementer_rekor_sign_generic(
        self,
        step_implementer,
        step_config={},
        step_name='',
        environment=None,
        implementer='',
        workflow_result=None,
        parent_work_dir_path='',
        artifact_to_sign_uri_config_key=None
    ):
        config = Config({
            Config.CONFIG_KEY: {
                step_name: [
                    {
                        'implementer': implementer,
                        'config': step_config
                    }
                ]

            }
        })
        step_config = config.get_step_config(step_name)
        sub_step_config = step_config.get_sub_step(implementer)

        if not workflow_result:
            workflow_result = WorkflowResult()

        step_implementer = step_implementer(
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path,
            config=sub_step_config,
            artifact_to_sign_uri_config_key=artifact_to_sign_uri_config_key,
            environment=environment
        )
        return step_implementer

    def test__required_config_or_result_keys(self):
        required_keys = RekorSignGeneric._required_config_or_result_keys()
        expected_required_keys = [
            'rekor-server-url',
            'signer-pgp-private-key'
        ]
        self.assertEqual(required_keys, expected_required_keys)

    @patch('ploigos_step_runner.step_implementers.shared.rekor_sign_generic.detach_sign_with_pgp_key')
    @patch('ploigos_step_runner.step_implementers.shared.rekor_sign_generic.base64_encode')
    def test_create_rekor_entry(self,
        base64_encode_mock,
        detach_sign_with_pgp_key_mock):
        with TempDirectory() as temp_dir:
            signer_pgp_public_key_path = os.path.join(
                os.path.dirname(__file__),
                '../../helpers','files',
                'ploigos-step-runner-tests-public.key'
            )

            #Copy key file to temp directory since _create_rekor_entry outputs a detached
            #signature to a new file.
            tmp_key_path = os.path.join(temp_dir.path, 'public.key')
            copy(signer_pgp_public_key_path, tmp_key_path)

            signer_pgp_private_key_fingerprint = self.TEST_FAKE_FINGERPRINT

            signer_pgp_public_key = self.TEST_FAKE_PUBLIC_KEY

            base64_encode_mock.return_value = "mock64"

            result = RekorSignGeneric._create_rekor_entry(self,
                signer_pgp_public_key,
                signer_pgp_private_key_fingerprint,
                path_to_file=tmp_key_path,
                artifact_to_sign_uri=self.TEST_ARTIFACT_TO_SIGN_URI
            )

            self.assertEqual(result['spec']['data']['hash']['value'], self.TEST_REKOR_ENTRY['spec']['data']['hash']['value'])
            self.assertEqual(result['spec']['extraData'], self.TEST_REKOR_ENTRY['spec']['extraData'])

            detach_sign_with_pgp_key_mock.assert_called_once_with(
                output_signature_path = tmp_key_path + '.asc',
                file_to_sign_path = tmp_key_path,
                pgp_private_key_fingerprint=signer_pgp_private_key_fingerprint
            )

            base64_encode_mock.assert_any_call(
                tmp_key_path + '.asc',
            )


    @patch.object(RekorSignGeneric, '_create_rekor_entry')
    @patch('sh.rekor', create=True)
    def test_upload_to_rekor(self, rekor_mock, create_mock):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            signer_pgp_public_key_path = os.path.join(
                os.path.dirname(__file__),
                '../../helpers','files',
                'ploigos-step-runner-tests-public.key'
            )

            step_config = {
                'rekor-server-url': self.TEST_REKOR_SERVER_URL,
                'signer-pgp-private-key': self.TEST_PGP_SIGNER_PRIVATE_KEY
            }

            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
                artifact_to_sign_uri_config_key=self.TEST_ARTIFACT_TO_SIGN_URI
            )

            artifact_data_file = os.path.join(parent_work_dir_path, 'automated-governance', 'automated-governance.json')
            artifact_data_file_path = Path(artifact_data_file)
            WorkflowResult().write_results_to_json_file(artifact_data_file_path)
            rekor_entry_path_name = os.path.join(parent_work_dir_path, 'automated-governance', 'entry.json')

            def create_mock_side_effect(signer_pgp_public_key_path, signer_pgp_private_key_user, extra_data_file):
                return self.TEST_REKOR_ENTRY

            def rekor_mock_side_effect(*args, **kwargs):
                return 'Created entry at: ' + args[2]+ '/api/v1/log/entries/' + self.TEST_REKOR_UUID

            create_mock.side_effect = create_mock_side_effect
            rekor_mock.side_effect = rekor_mock_side_effect

            rekor_entry_path = Path(rekor_entry_path_name)
            rekor_entry_path.touch()

            result_uuid = step_implementer._upload_to_rekor(
                rekor_server=self.TEST_REKOR_SERVER_URL,
                rekor_entry=self.TEST_REKOR_ENTRY
            )

            rekor_mock.assert_called_once_with(
                'upload',
                '--rekor_server',
                self.TEST_REKOR_SERVER_URL,
                '--entry',
                rekor_entry_path_name,
                _out=Any(IOBase),
                _err_to_out=True,
                _tee='out'
            )
            self.assertEqual(result_uuid, self.TEST_REKOR_UUID)
    @patch('ploigos_step_runner.step_implementers.shared.rekor_sign_generic.export_pgp_public_key')
    @patch('ploigos_step_runner.step_implementers.shared.rekor_sign_generic.import_pgp_key')
    @patch('ploigos_step_runner.step_implementers.shared.rekor_sign_generic.download_source_to_destination')
    @patch.object(RekorSignGeneric, '_upload_to_rekor')
    @patch.object(RekorSignGeneric, '_create_rekor_entry')
    def test__run_step(self, create_mock,
        upload_mock,
        download_source_to_destination_mock,
        import_pgp_key_mock,
        export_pgp_public_key_mock):

        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            signer_pgp_public_key_path = os.path.join(
                os.path.dirname(__file__),
                '../../helpers','files',
                'ploigos-step-runner-tests-public.key'
            )

            step_config = {
                'rekor-server-url': self.TEST_REKOR_SERVER_URL,
                'signer-pgp-private-key': self.TEST_FAKE_PRIVATE_KEY,
                self.TEST_ARTIFACT_TO_SIGN_URI_CONFIG_KEY: self.TEST_ARTIFACT_TO_SIGN_URI
            }

            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
                artifact_to_sign_uri_config_key=self.TEST_ARTIFACT_TO_SIGN_URI_CONFIG_KEY
            )

            expected_step_result = StepResult(
                step_name='automated-governance',
                sub_step_name='RekorSignGeneric',
                sub_step_implementer_name='RekorSignGeneric'
            )

            expected_step_result.add_artifact(
                name='rekor-entry',
                value=self.TEST_REKOR_ENTRY
            )
            expected_step_result.add_artifact(
                name='rekor-uuid',
                value=self.TEST_REKOR_UUID
            )
            rekor_uri = self.TEST_REKOR_ENTRY_URI
            expected_step_result.add_artifact(
                name='rekor-entry-uri',
                value=rekor_uri

            )

            download_source_to_destination_mock.return_value = signer_pgp_public_key_path
            import_pgp_key_mock.return_value = self.TEST_FAKE_FINGERPRINT
            export_pgp_public_key_mock.return_value = self.TEST_FAKE_PUBLIC_KEY

            create_mock.return_value = self.TEST_REKOR_ENTRY
            upload_mock.return_value = self.TEST_REKOR_UUID

            result = step_implementer._run_step()

            self.assertEqual(result, expected_step_result)



"""Configuration for Ploigos workflow.
"""

from ploigos_step_runner.config.config import Config
from ploigos_step_runner.config.config_value import ConfigValue
from ploigos_step_runner.config.config_value_decryptor import ConfigValueDecryptor
from ploigos_step_runner.config.step_config import StepConfig
from ploigos_step_runner.config.sub_step_config import SubStepConfig
